# ------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE.txt in the project root for
# license information.
# -------------------------------------------------------------------------
import asyncio
import os
from unittest import mock

import pytest
from devtools_testutils import (
    add_general_string_sanitizer,
    add_oauth_response_sanitizer,
    is_live,
    remove_batch_sanitizers,
    set_custom_default_matcher,
)
from azure.keyvault.keys._shared.client_base import DEFAULT_VERSION, ApiVersion

os.environ["PYTHONHASHSEED"] = "0"
ALL_API_VERSIONS = "--all-api-versions"


def pytest_addoption(parser):
    parser.addoption(
        ALL_API_VERSIONS,
        action="store_true",
        default=False,
        help="Test all api version in live mode. Not applicable in playback mode.",
    )


def pytest_configure(config):
    if is_live() and not config.getoption(ALL_API_VERSIONS):
        pytest.api_version = [DEFAULT_VERSION]
    else:
        pytest.api_version = ApiVersion


@pytest.fixture(scope="session", autouse=True)
def add_sanitizers(test_proxy):
    azure_keyvault_url = os.getenv("AZURE_KEYVAULT_URL", "https://vaultname.vault.azure.net")
    azure_keyvault_url = azure_keyvault_url.rstrip("/")
    keyvault_tenant_id = os.getenv("KEYVAULT_TENANT_ID", "keyvault_tenant_id")
    keyvault_subscription_id = os.getenv("KEYVAULT_SUBSCRIPTION_ID", "keyvault_subscription_id")
    azure_managedhsm_url = os.environ.get("AZURE_MANAGEDHSM_URL", "https://managedhsmvaultname.managedhsm.azure.net")
    azure_managedhsm_url = azure_managedhsm_url.rstrip("/")
    azure_attestation_uri = os.environ.get(
        "AZURE_KEYVAULT_ATTESTATION_URL", "https://fakeattestation.azurewebsites.net"
    )
    azure_attestation_uri = azure_attestation_uri.rstrip("/")

    add_general_string_sanitizer(target=azure_keyvault_url, value="https://vaultname.vault.azure.net")
    add_general_string_sanitizer(target=keyvault_tenant_id, value="00000000-0000-0000-0000-000000000000")
    add_general_string_sanitizer(target=keyvault_subscription_id, value="00000000-0000-0000-0000-000000000000")
    add_general_string_sanitizer(target=azure_managedhsm_url, value="https://managedhsmvaultname.managedhsm.azure.net")
    add_general_string_sanitizer(target=azure_attestation_uri, value="https://fakeattestation.azurewebsites.net")
    add_oauth_response_sanitizer()
    set_custom_default_matcher(compare_bodies=False, ignore_query_ordering=True, ignored_headers="Accept")

    # Remove the following sanitizers since certain fields are needed in tests and are non-sensitive:
    #  - AZSDK3430: $..id
    #  - AZSDK3447: $.key
    remove_batch_sanitizers(
        [
            "AZSDK3430",
            "AZSDK3447",
        ]
    )


@pytest.fixture(scope="session", autouse=True)
def patch_async_sleep():
    async def immediate_return(_):
        return

    if not is_live():
        with mock.patch("asyncio.sleep", immediate_return):
            yield

    else:
        yield


@pytest.fixture(scope="session", autouse=True)
def patch_sleep():
    def immediate_return(_):
        return

    if not is_live():
        with mock.patch("time.sleep", immediate_return):
            yield

    else:
        yield


@pytest.fixture(scope="session")
def event_loop(request):
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    yield loop
    loop.close()



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
from typing import Any, Callable, cast

from azure.core.exceptions import ResourceNotFoundError, HttpResponseError
from azure.core.pipeline import PipelineResponse
from azure.core.pipeline.transport import AsyncHttpTransport
from azure.core.polling import AsyncPollingMethod


class AsyncDeleteRecoverPollingMethod(AsyncPollingMethod):
    """Poller for deleting resources, and recovering deleted resources, in vaults with soft-delete enabled.

    This works by polling for the existence of the deleted or recovered resource. When a resource is deleted, Key Vault
    immediately removes it from its collection. However, the resource will not immediately appear in the deleted
    collection. Key Vault will therefore respond 404 to GET requests for the deleted resource; when it responds 2xx,
    the resource exists in the deleted collection i.e. its deletion is complete.

    Similarly, while recovering a deleted resource, Key Vault will respond 404 to GET requests for the non-deleted
    resource; when it responds 2xx, the resource exists in the non-deleted collection, i.e. its recovery is complete.

    :param pipeline_response: The operation's original pipeline response.
    :type pipeline_response: PipelineResponse
    :param command: An awaitable to invoke when polling.
    :type command: Callable
    :param final_resource: The final resource returned by the polling operation.
    :type final_resource: Any
    :param bool finished: Whether or not the polling operation is completed.
    :param int interval: The polling interval, in seconds.
    """

    def __init__(
        self,
        pipeline_response: PipelineResponse,
        command: Callable,
        final_resource: Any,
        finished: bool,
        interval: int = 2,
    ) -> None:
        self._pipeline_response = pipeline_response
        self._command = command
        self._resource = final_resource
        self._polling_interval = interval
        self._finished = finished

    def initialize(self, client, initial_response, deserialization_callback):
        pass

    async def _update_status(self) -> None:
        try:
            await self._command()
            self._finished = True
        except ResourceNotFoundError:
            pass
        except HttpResponseError as e:
            # If we are polling on get_deleted_* and we don't have get permissions, we will get
            # ResourceNotFoundError until the resource is recovered, at which point we'll get a 403.
            if e.status_code == 403:
                self._finished = True
            else:
                raise

    async def run(self) -> None:
        while not self.finished():
            await self._update_status()
            if not self.finished():
                # We should always ask the client's transport to sleep, instead of sleeping directly
                transport: AsyncHttpTransport = cast(AsyncHttpTransport, self._pipeline_response.context.transport)
                await transport.sleep(self._polling_interval)

    def finished(self) -> bool:
        return self._finished

    def resource(self) -> Any:
        return self._resource

    def status(self) -> str:
        return "finished" if self._finished else "polling"



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
from typing import TYPE_CHECKING

from .ec import EllipticCurveCryptographyProvider
from .local_provider import LocalCryptographyProvider
from .rsa import RsaCryptographyProvider
from .symmetric import SymmetricCryptographyProvider
from ... import KeyType

if TYPE_CHECKING:
    from ... import JsonWebKey


def get_local_cryptography_provider(key: "JsonWebKey") -> LocalCryptographyProvider:
    if key.kty in (KeyType.ec, KeyType.ec_hsm):  # type: ignore[attr-defined]
        return EllipticCurveCryptographyProvider(key)
    if key.kty in (KeyType.rsa, KeyType.rsa_hsm):  # type: ignore[attr-defined]
        return RsaCryptographyProvider(key)
    if key.kty in (KeyType.oct, KeyType.oct_hsm):  # type: ignore[attr-defined]
        return SymmetricCryptographyProvider(key)

    raise ValueError(f'Unsupported key type "{key.kty}"')  # type: ignore[attr-defined]


class NoLocalCryptography(LocalCryptographyProvider):
    def __init__(self):  # pylint:disable=super-init-not-called
        return

    def supports(self, operation, algorithm):
        return False

    def _get_internal_key(self, key):
        return None



# pylint: disable=line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
from __future__ import annotations
from typing import Any, cast, Optional, NoReturn, Union, TYPE_CHECKING

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric.padding import AsymmetricPadding, OAEP, PKCS1v15, PSS, MGF1
from cryptography.hazmat.primitives.asymmetric.rsa import (
    rsa_crt_dmp1,
    rsa_crt_dmq1,
    rsa_crt_iqmp,
    rsa_recover_prime_factors,
    RSAPrivateKey,
    RSAPrivateNumbers,
    RSAPublicKey,
    RSAPublicNumbers,
)
from cryptography.hazmat.primitives.asymmetric.utils import Prehashed
from cryptography.hazmat.primitives.hashes import Hash, HashAlgorithm, SHA1, SHA256, SHA384, SHA512
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    KeySerializationEncryption,
    PrivateFormat,
    PublicFormat,
)

from ._enums import EncryptionAlgorithm, KeyWrapAlgorithm, SignatureAlgorithm
from .._models import JsonWebKey

if TYPE_CHECKING:
    # Import client only during TYPE_CHECKING to avoid circular dependency
    from ._client import CryptographyClient


SIGN_ALGORITHM_MAP = {
    SHA256: SignatureAlgorithm.rs256,
    SHA384: SignatureAlgorithm.rs384,
    SHA512: SignatureAlgorithm.rs512,
}
OAEP_MAP = {SHA1: EncryptionAlgorithm.rsa_oaep, SHA256: EncryptionAlgorithm.rsa_oaep_256}
PSS_MAP = {
    SignatureAlgorithm.rs256: SignatureAlgorithm.ps256,
    SignatureAlgorithm.rs384: SignatureAlgorithm.ps384,
    SignatureAlgorithm.rs512: SignatureAlgorithm.ps512,
}


def get_encryption_algorithm(padding: AsymmetricPadding) -> EncryptionAlgorithm:
    """Maps an `AsymmetricPadding` to an encryption algorithm.

    :param padding: The padding to use.
    :type padding: ~cryptography.hazmat.primitives.asymmetric.padding.AsymmetricPadding

    :returns: The corresponding Key Vault encryption algorithm.
    :rtype: EncryptionAlgorithm
    """
    if isinstance(padding, OAEP):
        # Public algorithm property was only added in https://github.com/pyca/cryptography/pull/9582
        # _algorithm property has been available in every version of the OAEP class, so we use it as a backup
        try:
            algorithm = padding.algorithm  # type: ignore[attr-defined]
        except AttributeError:
            algorithm = padding._algorithm  # pylint:disable=protected-access
        mapped_algorithm = OAEP_MAP.get(type(algorithm))
        if mapped_algorithm is None:
            raise ValueError(f"Unsupported algorithm: {algorithm.name}")

        # Public mgf property was added at the same time as algorithm
        try:
            mgf = padding.mgf  # type: ignore[attr-defined]
        except AttributeError:
            mgf = padding._mgf  # pylint:disable=protected-access
        if not isinstance(mgf, MGF1):
            raise ValueError(f"Unsupported MGF: {mgf}")

    elif isinstance(padding, PKCS1v15):
        mapped_algorithm = EncryptionAlgorithm.rsa1_5
    else:
        raise ValueError(f"Unsupported padding: {padding.name}")

    return mapped_algorithm


def get_signature_algorithm(padding: AsymmetricPadding, algorithm: HashAlgorithm) -> SignatureAlgorithm:
    """Maps an `AsymmetricPadding` and `HashAlgorithm` to a signature algorithm.

    :param padding: The padding to use.
    :type padding: ~cryptography.hazmat.primitives.asymmetric.padding.AsymmetricPadding
    :param algorithm: The algorithm to use.
    :type algorithm: ~cryptography.hazmat.primitives.hashes.HashAlgorithm

    :returns: The corresponding Key Vault signature algorithm.
    :rtype: SignatureAlgorithm
    """
    mapped_algorithm = SIGN_ALGORITHM_MAP.get(type(algorithm))
    if mapped_algorithm is None:
        raise ValueError(f"Unsupported algorithm: {algorithm.name}")

    # If PSS padding is requested, use the PSS equivalent algorithm
    if isinstance(padding, PSS):
        mapped_algorithm = PSS_MAP.get(mapped_algorithm)

        # Public mgf property was only added in https://github.com/pyca/cryptography/pull/9582
        # _mgf property has been available in every version of the PSS class, so we use it as a backup
        try:
            mgf = padding.mgf  # type: ignore[attr-defined]
        except AttributeError:
            mgf = padding._mgf  # pylint:disable=protected-access
        if not isinstance(mgf, MGF1):
            raise ValueError(f"Unsupported MGF: {mgf}")

    # The only other padding accepted is PKCS1v15
    elif not isinstance(padding, PKCS1v15):
        raise ValueError(f"Unsupported padding: {padding.name}")

    return cast(SignatureAlgorithm, mapped_algorithm)


class KeyVaultRSAPublicKey(RSAPublicKey):
    """An `RSAPublicKey` implementation based on a key managed by Key Vault.

    This class should not be instantiated directly. Instead, use the
    :func:`~azure.keyvault.keys.crypto.CryptographyClient.create_rsa_public_key` method to create a key based on the
    client's key. Only synchronous clients and operations are supported at this time.
    """

    def __init__(self, client: "CryptographyClient", key_material: Optional[JsonWebKey] = None) -> None:
        self._client: "CryptographyClient" = client
        self._key: Optional[JsonWebKey] = key_material

    def encrypt(self, plaintext: bytes, padding: AsymmetricPadding) -> bytes:
        """Encrypts the given plaintext.

        :param bytes plaintext: Plaintext to encrypt.
        :param padding: The padding to use. Supported paddings are `OAEP` and `PKCS1v15`. For `OAEP` padding, supported
            hash algorithms are `SHA1` and `SHA256`. The only supported mask generation function is `MGF1`. See
            https://learn.microsoft.com/azure/key-vault/keys/about-keys-details for details.
        :type padding: ~cryptography.hazmat.primitives.asymmetric.padding.AsymmetricPadding

        :returns: The encrypted ciphertext, as bytes.
        :rtype: bytes
        """
        mapped_algorithm = get_encryption_algorithm(padding)
        result = self._client.encrypt(mapped_algorithm, plaintext)
        return result.ciphertext

    @property
    def key_size(self) -> int:
        """The bit length of the public modulus.

        :returns: The key's size.
        :rtype: int

        :raises ValueError: if the client is unable to obtain the key material from Key Vault.
        """
        if self._key is None:
            raise ValueError(
                "Key material could not be obtained from Key Vault. Only remote cryptographic operations "
                "(encrypt, verify) can be performed."
            )

        public_key = self.public_numbers().public_key()
        return public_key.key_size

    def public_numbers(self) -> RSAPublicNumbers:
        """Returns an `RSAPublicNumbers` representing the key's public numbers.

        :returns: The public numbers of the key.
        :rtype: RSAPublicNumbers

        :raises ValueError: if the client is unable to obtain the key material from Key Vault.
        """
        if self._key is None:
            raise ValueError(
                "Key material could not be obtained from Key Vault. Only remote cryptographic operations "
                "(encrypt, verify) can be performed."
            )

        e = int.from_bytes(self._key.e, "big")  # type: ignore[attr-defined]
        n = int.from_bytes(self._key.n, "big")  # type: ignore[attr-defined]
        return RSAPublicNumbers(e, n)

    def public_bytes(self, encoding: Encoding, format: PublicFormat) -> bytes:
        """Allows serialization of the key to bytes.

        This function uses the `cryptography` library's implementation.
        Encoding (`PEM` or `DER`) and format (`SubjectPublicKeyInfo` or `PKCS1`) are chosen to define the exact
        serialization.

        :param encoding: A value from the `Encoding` enum.
        :type encoding: ~cryptography.hazmat.primitives.serialization.Encoding
        :param format: A value from the `PublicFormat` enum.
        :type format: ~cryptography.hazmat.primitives.serialization.PublicFormat

        :returns: The serialized key.
        :rtype: bytes

        :raises ValueError: if the client is unable to obtain the key material from Key Vault.
        """
        if self._key is None:
            raise ValueError(
                "Key material could not be obtained from Key Vault. Only remote cryptographic operations "
                "(encrypt, verify) can be performed."
            )

        public_key = self.public_numbers().public_key()
        return public_key.public_bytes(encoding=encoding, format=format)

    def verify(
        self,
        signature: bytes,
        data: bytes,
        padding: AsymmetricPadding,
        algorithm: Union[Prehashed, HashAlgorithm],
    ) -> None:
        """Verifies the signature of the data.

        :param bytes signature: The signature to sign, as bytes.
        :param bytes data: The message string that was signed., as bytes.
        :param padding: The padding to use. Supported paddings are `PKCS1v15` and `PSS`. For `PSS`, the only supported
            mask generation function is `MGF1`. See https://learn.microsoft.com/azure/key-vault/keys/about-keys-details
            for details.
        :type padding: ~cryptography.hazmat.primitives.asymmetric.padding.AsymmetricPadding
        :param algorithm: The algorithm to sign with. Only `HashAlgorithm`s are supported -- specifically, `SHA256`,
            `SHA384`, and `SHA512`.
        :type algorithm: ~cryptography.hazmat.primitives.asymmetric.utils.Prehashed or
            cryptography.hazmat.primitives.hashes.HashAlgorithm

        :raises InvalidSignature: If the signature does not validate.
        """
        if isinstance(algorithm, Prehashed):
            raise ValueError("`Prehashed` algorithms are unsupported. Please provide a `HashAlgorithm` instead.")
        mapped_algorithm = get_signature_algorithm(padding, algorithm)
        digest = Hash(algorithm)
        digest.update(data)
        result = self._client.verify(mapped_algorithm, digest.finalize(), signature)
        if not result.is_valid:
            raise InvalidSignature(f"The provided signature '{signature!r}' is invalid.")

    def recover_data_from_signature(  # type: ignore[override]  # Parameter subset
        self,
        signature: bytes,
        padding: AsymmetricPadding,
        algorithm: Optional[HashAlgorithm],
    ) -> bytes:
        # pylint: disable=line-too-long
        """Recovers the signed data from the signature. Only supported with `cryptography` version 3.3 and above.

        This function uses the `cryptography` library's implementation.
        The data typically contains the digest of the original message string. The `padding` and `algorithm` parameters
        must match the ones used when the signature was created for the recovery to succeed.
        The `algorithm` parameter can also be set to None to recover all the data present in the signature, without
        regard to its format or the hash algorithm used for its creation.

        For `PKCS1v15` padding, this method returns the data after removing the padding layer. For standard signatures
        the data contains the full `DigestInfo` structure. For non-standard signatures, any data can be returned,
        including zero-length data.

        Normally you should use the `verify()` function to validate the signature. But for some non-standard signature
        formats you may need to explicitly recover and validate the signed data. The following are some examples:

        * Some old Thawte and Verisign timestamp certificates without `DigestInfo`.
        * Signed MD5/SHA1 hashes in TLS 1.1 or earlier (`RFC 4346 <https://datatracker.ietf.org/doc/html/rfc4346.html>`_, section 4.7).
        * IKE version 1 signatures without `DigestInfo` (`RFC 2409 <https://datatracker.ietf.org/doc/html/rfc2409.html>`_, section 5.1).

        :param bytes signature: The signature.
        :param padding: An instance of `AsymmetricPadding`. Recovery is only supported with some of the padding types.
        :type padding: ~cryptography.hazmat.primitives.asymmetric.padding.AsymmetricPadding
        :param algorithm: An instance of `HashAlgorithm`. Can be None to return all the data present in the signature.
        :type algorithm: ~cryptography.hazmat.primitives.hashes.HashAlgorithm

        :returns: The signed data.
        :rtype: bytes
        :raises NotImplementedError: if the local version of `cryptography` doesn't support this method.
        :raises ~cryptography.exceptions.InvalidSignature: if the signature is invalid.
        :raises ~cryptography.exceptions.UnsupportedAlgorithm: if the signature data recovery is not supported with
            the provided `padding` type.
        :raises ValueError: if the client is unable to obtain the key material from Key Vault.
        """
        if self._key is None:
            raise ValueError(
                "Key material could not be obtained from Key Vault. Only remote cryptographic operations "
                "(encrypt, verify) can be performed."
            )

        public_key = self.public_numbers().public_key()
        try:
            return public_key.recover_data_from_signature(signature=signature, padding=padding, algorithm=algorithm)
        except AttributeError as exc:
            raise NotImplementedError(
                "This method is only available on `cryptography`>=3.3. Update your package version to use this method."
            ) from exc

    def __eq__(self, other: object) -> bool:
        """Checks equality.

        :param object other: Another object to compare with this instance. Currently, only comparisons with
            `KeyVaultRSAPrivateKey` or `JsonWebKey` instances are supported.

        :returns: True if the objects are equal; False if the objects are unequal or if key material can't be obtained
            from Key Vault for comparison.
        :rtype: bool
        """
        if self._key is None:
            return False

        if isinstance(other, KeyVaultRSAPublicKey):
            return all(getattr(self._key, field) == getattr(other._key, field) for field in self._key._FIELDS)
        if isinstance(other, JsonWebKey):
            return all(getattr(self._key, field) == getattr(other, field) for field in self._key._FIELDS)
        return False

    def __copy__(self) -> KeyVaultRSAPublicKey:
        """Returns this instance since it is treated as immutable.

        :returns: This instance.
        :rtype: ~azure.keyvault.keys.crypto.KeyVaultRSAPublicKey
        """
        return self

    def __deepcopy__(self, memo: dict) -> KeyVaultRSAPublicKey:
        """Returns this instance since it is treated as immutable.

        :param dict memo: The memo dictionary used by deepcopy.
        :returns: This instance.
        :rtype: ~azure.keyvault.keys.crypto.KeyVaultRSAPublicKey
        """
        return self

    def verifier(  # pylint:disable=docstring-missing-param,docstring-missing-return,docstring-missing-rtype
        self, signature: bytes, padding: AsymmetricPadding, algorithm: HashAlgorithm
    ) -> NoReturn:
        """Not implemented. This method was deprecated in `cryptography` 2.0 and removed in 37.0.0."""
        raise NotImplementedError()


class KeyVaultRSAPrivateKey(RSAPrivateKey):
    """An `RSAPrivateKey` implementation based on a key managed by Key Vault.

    This class should not be instantiated directly. Instead, use the
    :func:`~azure.keyvault.keys.crypto.CryptographyClient.create_rsa_private_key` method to create a key based on the
    client's key. Only synchronous clients and operations are supported at this time.
    """

    def __init__(self, client: "CryptographyClient", key_material: Optional[JsonWebKey]) -> None:
        self._client: "CryptographyClient" = client
        self._key: Optional[JsonWebKey] = key_material

    def decrypt(self, ciphertext: bytes, padding: AsymmetricPadding) -> bytes:
        """Decrypts the provided ciphertext.

        :param bytes ciphertext: Encrypted bytes to decrypt.
        :param padding: The padding to use. Supported paddings are `OAEP` and `PKCS1v15`. For `OAEP` padding, supported
            hash algorithms are `SHA1` and `SHA256`. The only supported mask generation function is `MGF1`. See
            https://learn.microsoft.com/azure/key-vault/keys/about-keys-details for details.
        :type padding: ~cryptography.hazmat.primitives.asymmetric.padding.AsymmetricPadding

        :returns: The decrypted plaintext, as bytes.
        :rtype: bytes
        """
        mapped_algorithm = get_encryption_algorithm(padding)
        result = self._client.decrypt(mapped_algorithm, ciphertext)
        return result.plaintext

    @property
    def key_size(self) -> int:
        """The bit length of the public modulus.

        :returns: The key's size.
        :rtype: int

        :raises ValueError: if the client is unable to obtain the key material from Key Vault.
        """
        if self._key is None:
            raise ValueError(
                "Key material could not be obtained from Key Vault. Only remote cryptographic operations "
                "(decrypt, sign) can be performed."
            )

        # Key size only requires public modulus, which we can always get
        # Relying on private numbers instead would cause issues for keys stored in KV (which doesn't return private key)
        return self.public_key().key_size

    def public_key(self) -> KeyVaultRSAPublicKey:
        """The `RSAPublicKey` associated with this private key, as a `KeyVaultRSAPublicKey`.

        The public key implementation will use the same underlying cryptography client as this private key.

        :returns: The `KeyVaultRSAPublicKey` associated with the key.
        :rtype: ~azure.keyvault.keys.crypto.KeyVaultRSAPublicKey
        """
        return KeyVaultRSAPublicKey(self._client, self._key)

    def sign(  # type: ignore[override]  # Parameter subset
        self,
        data: bytes,
        padding: AsymmetricPadding,
        algorithm: Union[Prehashed, HashAlgorithm],
    ) -> bytes:
        """Signs the data.

        :param bytes data: The data to sign, as bytes.
        :param padding: The padding to use. Supported paddings are `PKCS1v15` and `PSS`. For `PSS`, the only supported
            mask generation function is `MGF1`. See https://learn.microsoft.com/azure/key-vault/keys/about-keys-details
            for details.
        :type padding: ~cryptography.hazmat.primitives.asymmetric.padding.AsymmetricPadding
        :param algorithm: The algorithm to sign with. Only `HashAlgorithm`s are supported -- specifically, `SHA256`,
            `SHA384`, and `SHA512`.
        :type algorithm: ~cryptography.hazmat.primitives.asymmetric.utils.Prehashed or
            cryptography.hazmat.primitives.hashes.HashAlgorithm

        :returns: The signature, as bytes.
        :rtype: bytes
        """
        if isinstance(algorithm, Prehashed):
            raise ValueError("`Prehashed` algorithms are unsupported. Please provide a `HashAlgorithm` instead.")
        mapped_algorithm = get_signature_algorithm(padding, algorithm)
        digest = Hash(algorithm)
        digest.update(data)
        result = self._client.sign(mapped_algorithm, digest.finalize())
        return result.signature

    def private_numbers(self) -> RSAPrivateNumbers:
        """Returns an `RSAPrivateNumbers` representing the key's private numbers.

        :returns: The private numbers of the key.
        :rtype: ~cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateNumbers

        :raises ValueError: if the client is unable to obtain the key material from Key Vault.
        """
        if self._key is None:
            raise ValueError(
                "Key material could not be obtained from Key Vault. Only remote cryptographic operations "
                "(decrypt, sign) can be performed."
            )

        # Fetch public numbers from JWK
        e = int.from_bytes(self._key.e, "big")  # type: ignore[attr-defined]
        n = int.from_bytes(self._key.n, "big")  # type: ignore[attr-defined]
        public_numbers = RSAPublicNumbers(e, n)

        # Fetch private numbers from JWK
        p = int.from_bytes(self._key.p, "big") if self._key.p else None  # type: ignore[attr-defined]
        q = int.from_bytes(self._key.q, "big") if self._key.q else None  # type: ignore[attr-defined]
        d = int.from_bytes(self._key.d, "big") if self._key.d else None  # type: ignore[attr-defined]
        dmp1 = int.from_bytes(self._key.dp, "big") if self._key.dp else None  # type: ignore[attr-defined]
        dmq1 = int.from_bytes(self._key.dq, "big") if self._key.dq else None  # type: ignore[attr-defined]
        iqmp = int.from_bytes(self._key.qi, "big") if self._key.qi else None  # type: ignore[attr-defined]

        # Calculate any missing attributes
        if d is None:
            raise ValueError("An 'RSAPrivateNumbers' couldn't be created with the available key material.")
        if p is None or q is None:
            p, q = rsa_recover_prime_factors(n, e, d)
        if dmp1 is None:
            dmp1 = rsa_crt_dmp1(d, p)
        if dmq1 is None:
            dmq1 = rsa_crt_dmq1(d, q)
        if iqmp is None:
            iqmp = rsa_crt_iqmp(p, q)

        return RSAPrivateNumbers(p, q, d, dmp1, dmq1, iqmp, public_numbers)

    def private_bytes(
        self, encoding: Encoding, format: PrivateFormat, encryption_algorithm: KeySerializationEncryption
    ) -> bytes:
        """Allows serialization of the key to bytes.

        This function uses the `cryptography` library's implementation.
        Encoding (`PEM` or `DER`) and format (`TraditionalOpenSSL`, `OpenSSH`, or `PKCS8`) and encryption algorithm
        (such as `BestAvailableEncryption` or `NoEncryption`) are chosen to define the exact serialization.

        :param encoding: A value from the `Encoding` enum.
        :type encoding: ~cryptography.hazmat.primitives.serialization.Encoding
        :param format: A value from the `PrivateFormat` enum.
        :type format: ~cryptography.hazmat.primitives.serialization.PrivateFormat
        :param encryption_algorithm: An instance of an object conforming to the `KeySerializationEncryption` interface.
        :type encryption_algorithm: ~cryptography.hazmat.primitives.serialization.KeySerializationEncryption

        :returns: The serialized key.
        :rtype: bytes

        :raises ValueError: if the client is unable to obtain the key material from Key Vault.
        """
        if self._key is None:
            raise ValueError(
                "Key material could not be obtained from Key Vault. Only remote cryptographic operations "
                "(decrypt, sign) can be performed."
            )

        try:
            private_numbers = self.private_numbers()
        except ValueError as exc:
            raise ValueError("Insufficient key material to serialize the private key.") from exc
        private_key = private_numbers.private_key()
        return private_key.private_bytes(encoding=encoding, format=format, encryption_algorithm=encryption_algorithm)

    def signer(  # pylint:disable=docstring-missing-param,docstring-missing-return,docstring-missing-rtype
        self, padding: AsymmetricPadding, algorithm: HashAlgorithm
    ) -> NoReturn:
        """Not implemented. This method was deprecated in `cryptography` 2.0 and removed in 37.0.0."""
        raise NotImplementedError()

    def __copy__(self) -> KeyVaultRSAPrivateKey:
        """Returns this instance since it is treated as immutable.

        :returns: This instance.
        :rtype: ~azure.keyvault.keys.crypto.KeyVaultRSAPrivateKey
        """
        return self

    def __deepcopy__(self, memo: dict) -> KeyVaultRSAPrivateKey:
        """Returns this instance since it is treated as immutable.

        :param dict memo: The memo dictionary used by deepcopy.
        :returns: This instance.
        :rtype: ~azure.keyvault.keys.crypto.KeyVaultRSAPrivateKey
        """
        return self


class DecryptResult:
    """The result of a decrypt operation.

    :param str key_id: The encryption key's Key Vault identifier
    :param algorithm: The encryption algorithm used
    :type algorithm: ~azure.keyvault.keys.crypto.EncryptionAlgorithm
    :param bytes plaintext: The decrypted bytes
    """

    def __init__(self, key_id: Optional[str], algorithm: EncryptionAlgorithm, plaintext: bytes) -> None:
        self.key_id = key_id
        self.algorithm = algorithm
        self.plaintext = plaintext


class EncryptResult:
    """The result of an encrypt operation.

    :param str key_id: The encryption key's Key Vault identifier
    :param algorithm: The encryption algorithm used
    :type algorithm: ~azure.keyvault.keys.crypto.EncryptionAlgorithm
    :param bytes ciphertext: The encrypted bytes

    :keyword bytes iv: Initialization vector for symmetric algorithms
    :keyword bytes authentication_tag: The tag to authenticate when performing decryption with an authenticated
        algorithm
    :keyword bytes additional_authenticated_data: Additional data to authenticate but not encrypt/decrypt when using an
        authenticated algorithm
    """

    def __init__(self, key_id: Optional[str], algorithm: EncryptionAlgorithm, ciphertext: bytes, **kwargs: Any) -> None:
        self.key_id = key_id
        self.algorithm = algorithm
        self.ciphertext = ciphertext
        self.iv = kwargs.pop("iv", None)
        self.tag = kwargs.pop("authentication_tag", None)
        self.aad = kwargs.pop("additional_authenticated_data", None)


class SignResult:
    """The result of a sign operation.

    :param str key_id: The signing key's Key Vault identifier
    :param algorithm: The signature algorithm used
    :type algorithm: ~azure.keyvault.keys.crypto.SignatureAlgorithm
    :param bytes signature:
    """

    def __init__(self, key_id: Optional[str], algorithm: SignatureAlgorithm, signature: bytes) -> None:
        self.key_id = key_id
        self.algorithm = algorithm
        self.signature = signature


class VerifyResult:
    """The result of a verify operation.

    :param str key_id: The signing key's Key Vault identifier
    :param bool is_valid: Whether the signature is valid
    :param algorithm: The signature algorithm used
    :type algorithm: ~azure.keyvault.keys.crypto.SignatureAlgorithm
    """

    def __init__(self, key_id: Optional[str], is_valid: bool, algorithm: SignatureAlgorithm) -> None:
        self.key_id = key_id
        self.is_valid = is_valid
        self.algorithm = algorithm


class UnwrapResult:
    """The result of an unwrap key operation.

    :param str key_id: Key encryption key's Key Vault identifier
    :param algorithm: The key wrap algorithm used
    :type algorithm: ~azure.keyvault.keys.crypto.KeyWrapAlgorithm
    :param bytes key: The unwrapped key
    """

    def __init__(self, key_id: Optional[str], algorithm: KeyWrapAlgorithm, key: bytes) -> None:
        self.key_id = key_id
        self.algorithm = algorithm
        self.key = key


class WrapResult:
    """The result of a wrap key operation.

    :param str key_id: The wrapping key's Key Vault identifier
    :param algorithm: The key wrap algorithm used
    :type algorithm: ~azure.keyvault.keys.crypto.KeyWrapAlgorithm
    :param bytes encrypted_key: The encrypted key bytes
    """

    def __init__(self, key_id: Optional[str], algorithm: KeyWrapAlgorithm, encrypted_key: bytes) -> None:
        self.key_id = key_id
        self.algorithm = algorithm
        self.encrypted_key = encrypted_key



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# -------------------------------------
from ._enums import KeyCurveName, KeyExportEncryptionAlgorithm, KeyOperation, KeyRotationPolicyAction, KeyType
from ._shared.client_base import ApiVersion
from ._models import (
    DeletedKey,
    JsonWebKey,
    KeyAttestation,
    KeyProperties,
    KeyReleasePolicy,
    KeyRotationLifetimeAction,
    KeyRotationPolicy,
    KeyVaultKey,
    KeyVaultKeyIdentifier,
    ReleaseKeyResult,
)
from ._client import KeyClient

__all__ = [
    "ApiVersion",
    "KeyClient",
    "JsonWebKey",
    "KeyAttestation",
    "KeyVaultKey",
    "KeyVaultKeyIdentifier",
    "KeyCurveName",
    "KeyExportEncryptionAlgorithm",
    "KeyOperation",
    "KeyRotationPolicyAction",
    "KeyType",
    "DeletedKey",
    "KeyProperties",
    "KeyReleasePolicy",
    "KeyRotationLifetimeAction",
    "KeyRotationPolicy",
    "ReleaseKeyResult",
]

from ._version import VERSION

__version__ = VERSION



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------
# pylint: disable=wrong-import-position

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ._patch import *  # pylint: disable=unused-wildcard-import


from ._models import (  # type: ignore
    BackupKeyResult,
    DeletedKeyBundle,
    DeletedKeyItem,
    GetRandomBytesRequest,
    JsonWebKey,
    KeyAttestation,
    KeyAttributes,
    KeyBundle,
    KeyCreateParameters,
    KeyImportParameters,
    KeyItem,
    KeyOperationResult,
    KeyOperationsParameters,
    KeyReleaseParameters,
    KeyReleasePolicy,
    KeyReleaseResult,
    KeyRestoreParameters,
    KeyRotationPolicy,
    KeyRotationPolicyAttributes,
    KeySignParameters,
    KeyUpdateParameters,
    KeyVaultError,
    KeyVaultErrorError,
    KeyVerifyParameters,
    KeyVerifyResult,
    LifetimeActions,
    LifetimeActionsTrigger,
    LifetimeActionsType,
    RandomBytes,
)

from ._enums import (  # type: ignore
    DeletionRecoveryLevel,
    JsonWebKeyCurveName,
    JsonWebKeyEncryptionAlgorithm,
    JsonWebKeyOperation,
    JsonWebKeySignatureAlgorithm,
    JsonWebKeyType,
    KeyEncryptionAlgorithm,
    KeyRotationPolicyAction,
)
from ._patch import __all__ as _patch_all
from ._patch import *
from ._patch import patch_sdk as _patch_sdk

__all__ = [
    "BackupKeyResult",
    "DeletedKeyBundle",
    "DeletedKeyItem",
    "GetRandomBytesRequest",
    "JsonWebKey",
    "KeyAttestation",
    "KeyAttributes",
    "KeyBundle",
    "KeyCreateParameters",
    "KeyImportParameters",
    "KeyItem",
    "KeyOperationResult",
    "KeyOperationsParameters",
    "KeyReleaseParameters",
    "KeyReleasePolicy",
    "KeyReleaseResult",
    "KeyRestoreParameters",
    "KeyRotationPolicy",
    "KeyRotationPolicyAttributes",
    "KeySignParameters",
    "KeyUpdateParameters",
    "KeyVaultError",
    "KeyVaultErrorError",
    "KeyVerifyParameters",
    "KeyVerifyResult",
    "LifetimeActions",
    "LifetimeActionsTrigger",
    "LifetimeActionsType",
    "RandomBytes",
    "DeletionRecoveryLevel",
    "JsonWebKeyCurveName",
    "JsonWebKeyEncryptionAlgorithm",
    "JsonWebKeyOperation",
    "JsonWebKeySignatureAlgorithm",
    "JsonWebKeyType",
    "KeyEncryptionAlgorithm",
    "KeyRotationPolicyAction",
]
__all__.extend([p for p in _patch_all if p not in __all__])  # pyright: ignore
_patch_sdk()



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import abc
import os
from typing import Optional, TYPE_CHECKING, Union

from azure.core.exceptions import AzureError

from .. import DecryptResult, EncryptResult, SignResult, UnwrapResult, VerifyResult, WrapResult
from ... import KeyOperation

ABC = abc.ABC

if TYPE_CHECKING:
    from .._internal.key import Key
    from .. import EncryptionAlgorithm, KeyWrapAlgorithm, SignatureAlgorithm
    from ... import JsonWebKey

    Algorithm = Union[EncryptionAlgorithm, KeyWrapAlgorithm, SignatureAlgorithm]


class LocalCryptographyProvider(ABC):
    def __init__(self, key: "JsonWebKey") -> None:
        self._allowed_ops = frozenset(key.key_ops or [])  # type: ignore[attr-defined]
        self._internal_key = self._get_internal_key(key)
        self._key = key

    @abc.abstractmethod
    def _get_internal_key(self, key: "JsonWebKey") -> "Key":
        pass

    @abc.abstractmethod
    def supports(self, operation: KeyOperation, algorithm: "Algorithm") -> bool:
        pass

    @property
    def key_id(self) -> "Optional[str]":
        """The full identifier of the provider's key.

        :returns: The full identifier of the provider's key.
        :rtype: str or None
        """
        return self._key.kid  # type: ignore[attr-defined]

    def _raise_if_unsupported(self, operation: KeyOperation, algorithm: "Algorithm") -> None:
        if not self.supports(operation, algorithm):
            raise NotImplementedError(
                f'This key does not support the "{operation}" operation with algorithm "{algorithm}"'
            )
        if operation not in self._allowed_ops:
            raise AzureError(f'This key does not allow the "{operation}" operation')

    def encrypt(
        self, algorithm: "EncryptionAlgorithm", plaintext: bytes, iv: "Optional[bytes]" = None
    ) -> EncryptResult:
        self._raise_if_unsupported(KeyOperation.encrypt, algorithm)

        # If an IV isn't provided with AES-CBCPAD encryption, try to create one
        if iv is None and algorithm.value.endswith("CBCPAD"):
            try:
                iv = os.urandom(16)
            except NotImplementedError as ex:
                raise ValueError(
                    "An IV could not be generated on this OS. Please provide your own cryptographically random, "
                    "non-repeating IV for local cryptography."
                ) from ex

        ciphertext = self._internal_key.encrypt(plaintext, algorithm=algorithm.value, iv=iv)
        return EncryptResult(
            key_id=self._key.kid, algorithm=algorithm, ciphertext=ciphertext, iv=iv  # type: ignore[attr-defined]
        )

    def decrypt(
        self, algorithm: "EncryptionAlgorithm", ciphertext: bytes, iv: "Optional[bytes]" = None
    ) -> DecryptResult:
        self._raise_if_unsupported(KeyOperation.decrypt, algorithm)
        plaintext = self._internal_key.decrypt(ciphertext, iv=iv, algorithm=algorithm.value)
        return DecryptResult(
            key_id=self._key.kid, algorithm=algorithm, plaintext=plaintext  # type: ignore[attr-defined]
        )

    def wrap_key(self, algorithm: "KeyWrapAlgorithm", key: bytes) -> "WrapResult":
        self._raise_if_unsupported(KeyOperation.wrap_key, algorithm)
        encrypted_key = self._internal_key.wrap_key(key, algorithm=algorithm.value)
        return WrapResult(
            key_id=self._key.kid, algorithm=algorithm, encrypted_key=encrypted_key  # type: ignore[attr-defined]
        )

    def unwrap_key(self, algorithm: "KeyWrapAlgorithm", encrypted_key: bytes) -> "UnwrapResult":
        self._raise_if_unsupported(KeyOperation.unwrap_key, algorithm)
        unwrapped_key = self._internal_key.unwrap_key(encrypted_key, algorithm=algorithm.value)
        return UnwrapResult(key_id=self._key.kid, algorithm=algorithm, key=unwrapped_key)  # type: ignore[attr-defined]

    def sign(self, algorithm: "SignatureAlgorithm", digest: bytes) -> "SignResult":
        self._raise_if_unsupported(KeyOperation.sign, algorithm)
        signature = self._internal_key.sign(digest, algorithm=algorithm.value)
        return SignResult(key_id=self._key.kid, algorithm=algorithm, signature=signature)  # type: ignore[attr-defined]

    def verify(self, algorithm: "SignatureAlgorithm", digest: bytes, signature: bytes) -> "VerifyResult":
        self._raise_if_unsupported(KeyOperation.verify, algorithm)
        is_valid = self._internal_key.verify(digest, signature, algorithm=algorithm.value)
        return VerifyResult(key_id=self._key.kid, algorithm=algorithm, is_valid=is_valid)  # type: ignore[attr-defined]



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import json
import os

import pytest
from azure.core.pipeline import AsyncPipeline
from azure.core.pipeline.transport import AioHttpTransport, HttpRequest
from devtools_testutils import AzureRecordedTestCase


async def get_attestation_token(attestation_uri):
    request = HttpRequest("GET", f"{attestation_uri}/generate-test-token")
    async with AsyncPipeline(transport=AioHttpTransport()) as pipeline:
        response = await pipeline.run(request)
        return json.loads(response.http_response.text())["token"]


class AsyncKeysClientPreparer(AzureRecordedTestCase):
    def __init__(self, *args, **kwargs):
        vault_playback_url = "https://vaultname.vault.azure.net"
        hsm_playback_url = "https://managedhsmvaultname.managedhsm.azure.net"
        self.is_logging_enabled = kwargs.pop("logging_enable", True)

        if self.is_live:
            self.vault_url = os.environ["AZURE_KEYVAULT_URL"]
            hsm = os.environ.get("AZURE_MANAGEDHSM_URL")
            self.managed_hsm_url = hsm if hsm else None
        else:
            self.vault_url = vault_playback_url
            self.managed_hsm_url = hsm_playback_url

        self._set_mgmt_settings_real_values()

    def __call__(self, fn):
        async def _preparer(test_class, api_version, is_hsm, **kwargs):

            self._skip_if_not_configured(is_hsm)
            if not self.is_logging_enabled:
                kwargs.update({"logging_enable": False})
            endpoint_url = self.managed_hsm_url if is_hsm else self.vault_url
            client = self.create_key_client(endpoint_url, api_version=api_version, **kwargs)
            async with client:
                await fn(
                    test_class, client, is_hsm=is_hsm, managed_hsm_url=self.managed_hsm_url, vault_url=self.vault_url
                )

        return _preparer

    def create_key_client(self, vault_uri, **kwargs):

        from azure.keyvault.keys.aio import KeyClient

        credential = self.get_credential(KeyClient, is_async=True)

        return self.create_client_from_credential(KeyClient, credential=credential, vault_url=vault_uri, **kwargs)

    def _set_mgmt_settings_real_values(self):
        if self.is_live:
            os.environ["AZURE_TENANT_ID"] = os.getenv("KEYVAULT_TENANT_ID", "")  # empty in pipelines
            os.environ["AZURE_CLIENT_ID"] = os.getenv("KEYVAULT_CLIENT_ID", "")  # empty in pipelines
            os.environ["AZURE_CLIENT_SECRET"] = os.getenv("KEYVAULT_CLIENT_SECRET", "")  # empty for user-based auth

    def _skip_if_not_configured(self, is_hsm):
        if self.is_live and is_hsm and self.managed_hsm_url is None:
            pytest.skip("No HSM endpoint for live testing")



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import threading
from typing import Dict, Optional
from urllib import parse

from .http_challenge import HttpChallenge


_cache: "Dict[str, HttpChallenge]" = {}
_lock = threading.Lock()


def get_challenge_for_url(url: str) -> "Optional[HttpChallenge]":
    """Gets the challenge for the cached URL.

    :param str url: the URL the challenge is cached for.

    :returns: The challenge for the cached request URL, or None if the request URL isn't cached.
    :rtype: HttpChallenge or None
    """

    if not url:
        raise ValueError("URL cannot be None")

    key = _get_cache_key(url)

    with _lock:
        return _cache.get(key.lower())


def _get_cache_key(url: str) -> str:
    """Use the URL's netloc as cache key except when the URL specifies the default port for its scheme. In that case
    use the netloc without the port. That is to say, https://foo.bar and https://foo.bar:443 are considered equivalent.

    This equivalency prevents an unnecessary challenge when using Key Vault's paging API. The Key Vault client doesn't
    specify ports, but Key Vault's next page links do, so a redundant challenge would otherwise be executed when the
    client requests the next page.

    :param str url: The HTTP request URL.

    :returns: The URL's `netloc`, minus any port attached to the URL.
    :rtype: str
    """

    parsed = parse.urlparse(url)
    if parsed.scheme == "https" and parsed.port == 443:
        return parsed.netloc[:-4]
    return parsed.netloc


def remove_challenge_for_url(url: str) -> None:
    """Removes the cached challenge for the specified URL.

    :param str url: the URL for which to remove the cached challenge
    """
    if not url:
        raise ValueError("URL cannot be empty")

    key = _get_cache_key(url)
    with _lock:
        del _cache[key.lower()]


def set_challenge_for_url(url: str, challenge: "HttpChallenge") -> None:
    """Caches the challenge for the specified URL.

    :param str url: the URL for which to cache the challenge
    :param challenge: the challenge to cache
    :type challenge: HttpChallenge
    """
    if not url:
        raise ValueError("URL cannot be empty")

    if not challenge:
        raise ValueError("Challenge cannot be empty")

    src_url = parse.urlparse(url)
    if src_url.netloc.lower() != challenge.source_authority.lower():
        raise ValueError("Source URL and Challenge URL do not match")

    key = _get_cache_key(url)
    with _lock:
        _cache[key.lower()] = challenge


def clear() -> None:
    """Clears the cache."""

    with _lock:
        _cache.clear()



# pylint: disable=line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import asyncio
import os
from azure.identity.aio import DefaultAzureCredential
from azure.keyvault.keys.aio import KeyClient

# ----------------------------------------------------------------------------------------------------------
# Prerequisites:
# 1. An Azure Key Vault (https://learn.microsoft.com/azure/key-vault/quick-create-cli)
#
# 2. azure-keyvault-keys and azure-identity libraries (pip install these)
#
# 3. Set environment variable VAULT_URL with the URL of your key vault
#
# 4. Set up your environment to use azure-identity's DefaultAzureCredential. For more information about how to configure
#    the DefaultAzureCredential, refer to https://aka.ms/azsdk/python/identity/docs#azure.identity.DefaultAzureCredential
#
# 5. Key create, backup, delete, purge, and restore permissions for your service principal in your vault
#
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates the basic backup and restore operations on a vault(key) resource for Azure Key Vault
#
# 1. Create a key (create_key)
#
# 2. Backup a key (backup_key)
#
# 3. Delete a key (delete_key)
#
# 4. Purge a key (purge_deleted_key)
#
# 5. Restore a key (restore_key_backup)
# ----------------------------------------------------------------------------------------------------------


async def run_sample():
    # Instantiate a key client that will be used to call the service.
    # Here we use the DefaultAzureCredential, but any azure-identity credential can be used.
    VAULT_URL = os.environ["VAULT_URL"]
    credential = DefaultAzureCredential()
    client = KeyClient(vault_url=VAULT_URL, credential=credential)

    # Let's create a Key of type RSA.
    # if the key already exists in the Key Vault, then a new version of the key is created.
    print("\n.. Create Key")
    key = await client.create_key("keyNameAsync", "RSA")
    print(f"Key with name '{key.name}' created with key type '{key.key_type}'")

    # Backups are good to have, if in case keys gets deleted accidentally.
    # For long term storage, it is ideal to write the backup to a file.
    print("\n.. Create a backup for an existing Key")
    key_backup = await client.backup_key(key.name)
    print(f"Backup created for key with name '{key.name}'.")

    # The rsa key is no longer in use, so you delete it.
    deleted_key = await client.delete_key(key.name)
    print(f"Deleted key with name '{deleted_key.name}'")

    # Purge the deleted key.
    # The purge will take some time, so wait before restoring the backup to avoid a conflict.
    print("\n.. Purge the key")
    await client.purge_deleted_key(key.name)
    await asyncio.sleep(60)
    print(f"Purged key with name '{deleted_key.name}'")

    # In the future, if the key is required again, we can use the backup value to restore it in the Key Vault.
    print("\n.. Restore the key using the backed up key bytes")
    key = await client.restore_key_backup(key_backup)
    print(f"Restored key with name '{key.name}'")

    print("\nrun_sample done")
    await credential.close()
    await client.close()


if __name__ == "__main__":
    asyncio.run(run_sample())



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import hashlib
import os

import pytest
from azure.keyvault.keys import KeyCurveName, KeyVaultKey
from azure.keyvault.keys.crypto import EncryptionAlgorithm, KeyWrapAlgorithm, SignatureAlgorithm
from azure.keyvault.keys.crypto._providers import get_local_cryptography_provider

from keys import EC_KEYS, RSA_KEYS


@pytest.mark.parametrize(
    "key,algorithm,hash_function",
    (
        (EC_KEYS[KeyCurveName.p_256], SignatureAlgorithm.es256, hashlib.sha256),
        (EC_KEYS[KeyCurveName.p_256_k], SignatureAlgorithm.es256_k, hashlib.sha256),
        (EC_KEYS[KeyCurveName.p_384], SignatureAlgorithm.es384, hashlib.sha384),
        (EC_KEYS[KeyCurveName.p_521], SignatureAlgorithm.es512, hashlib.sha512),
    ),
)
def test_ec_sign_verify(key, algorithm, hash_function):
    provider = get_local_cryptography_provider(key.key)
    digest = hash_function(b"message").digest()
    sign_result = provider.sign(algorithm, digest)
    verify_result = provider.verify(sign_result.algorithm, digest, sign_result.signature)
    assert verify_result.is_valid


@pytest.mark.parametrize("key", RSA_KEYS.values())
@pytest.mark.parametrize("algorithm", (a for a in EncryptionAlgorithm if a.startswith("RSA")))
def test_rsa_encrypt_decrypt(key, algorithm):
    provider = get_local_cryptography_provider(key.key)
    plaintext = b"plaintext"
    encrypt_result = provider.encrypt(algorithm, plaintext)
    decrypt_result = provider.decrypt(encrypt_result.algorithm, encrypt_result.ciphertext)
    assert decrypt_result.plaintext == plaintext


@pytest.mark.parametrize(
    "algorithm,key_size",
    (
        (EncryptionAlgorithm.a256_cbcpad, 32),
        (EncryptionAlgorithm.a192_cbcpad, 24),
        (EncryptionAlgorithm.a128_cbcpad, 16),
    ),
)
def test_symmetric_encrypt_decrypt(algorithm, key_size):
    jwk = {
        "k": os.urandom(key_size),
        "kid": "http://localhost/keys/key/version",
        "kty": "oct-HSM",
        "key_ops": ("encrypt", "decrypt"),
    }
    key = KeyVaultKey(key_id="http://localhost/keys/key/version", jwk=jwk)
    provider = get_local_cryptography_provider(key.key)
    plaintext = b"plaintext"
    iv = os.urandom(16)

    encrypt_result = provider.encrypt(algorithm, plaintext, iv=iv)
    assert encrypt_result.key_id == key.id
    assert encrypt_result.iv == iv

    decrypt_result = provider.decrypt(encrypt_result.algorithm, encrypt_result.ciphertext, iv=encrypt_result.iv)
    assert decrypt_result.plaintext == plaintext

    # Try with no IV to verify that we generate one for local crypto
    encrypt_result = provider.encrypt(algorithm, plaintext, iv=None)
    assert encrypt_result.key_id == key.id
    assert len(encrypt_result.iv) == 16

    decrypt_result = provider.decrypt(encrypt_result.algorithm, encrypt_result.ciphertext, iv=encrypt_result.iv)
    assert decrypt_result.plaintext == plaintext


@pytest.mark.parametrize("key", RSA_KEYS.values())
@pytest.mark.parametrize(
    "algorithm,hash_function",
    (
        (
            (SignatureAlgorithm.ps256, hashlib.sha256),
            (SignatureAlgorithm.ps384, hashlib.sha384),
            (SignatureAlgorithm.ps512, hashlib.sha512),
            (SignatureAlgorithm.rs256, hashlib.sha256),
            (SignatureAlgorithm.rs384, hashlib.sha384),
            (SignatureAlgorithm.rs512, hashlib.sha512),
        )
    ),
)
def test_rsa_sign_verify(key, algorithm, hash_function):
    message = b"message"
    provider = get_local_cryptography_provider(key.key)
    digest = hash_function(message).digest()
    sign_result = provider.sign(algorithm, digest)
    verify_result = provider.verify(sign_result.algorithm, digest, sign_result.signature)
    assert verify_result.is_valid


@pytest.mark.parametrize("key", RSA_KEYS.values())
@pytest.mark.parametrize("algorithm", (a for a in KeyWrapAlgorithm if a.startswith("RSA")))
def test_rsa_wrap_unwrap(key, algorithm):
    plaintext = b"arbitrary bytes"
    key.key.kid = key.id
    provider = get_local_cryptography_provider(key.key)

    wrap_result = provider.wrap_key(algorithm, plaintext)
    assert wrap_result.key_id == key.id

    unwrap_result = provider.unwrap_key(wrap_result.algorithm, wrap_result.encrypted_key)
    assert unwrap_result.key == plaintext


@pytest.mark.parametrize("algorithm", (a for a in KeyWrapAlgorithm if a.startswith("A")))
def test_symmetric_wrap_unwrap(algorithm):
    jwk = {
        "k": os.urandom(32),
        "kid": "http://localhost/keys/key/version",
        "kty": "oct",
        "key_ops": ("unwrapKey", "wrapKey"),
    }
    key = KeyVaultKey(key_id="http://localhost/keys/key/version", jwk=jwk)
    provider = get_local_cryptography_provider(key.key)
    key_bytes = os.urandom(32)

    wrap_result = provider.wrap_key(algorithm, key_bytes)
    assert wrap_result.key_id == key.id

    unwrap_result = provider.unwrap_key(wrap_result.algorithm, wrap_result.encrypted_key)
    assert unwrap_result.key == key_bytes


@pytest.mark.parametrize("key", RSA_KEYS.values())
@pytest.mark.parametrize(
    "algorithm",
    [a for a in KeyWrapAlgorithm if not a.startswith("RSA")] + [a for a in SignatureAlgorithm if a.startswith("ES")],
)
def test_unsupported_rsa_operations(key, algorithm):
    """The crypto provider should raise NotImplementedError when a key doesn't support an operation or algorithm"""

    provider = get_local_cryptography_provider(key.key)
    if isinstance(algorithm, EncryptionAlgorithm):
        with pytest.raises(NotImplementedError):
            provider.encrypt(algorithm, b"...")
        with pytest.raises(NotImplementedError):
            provider.decrypt(algorithm, b"...")
    if isinstance(algorithm, KeyWrapAlgorithm):
        with pytest.raises(NotImplementedError):
            provider.wrap_key(algorithm, b"...")
        with pytest.raises(NotImplementedError):
            provider.unwrap_key(algorithm, b"...")
    elif isinstance(algorithm, SignatureAlgorithm):
        with pytest.raises(NotImplementedError):
            provider.sign(algorithm, b"...")
        with pytest.raises(NotImplementedError):
            provider.verify(algorithm, b"...", b"...")


@pytest.mark.parametrize("key", EC_KEYS.values())
@pytest.mark.parametrize(
    "algorithm",
    [a for a in KeyWrapAlgorithm if a.startswith("RSA")]
    + [a for a in SignatureAlgorithm if not a.startswith("ES")]
    + list(EncryptionAlgorithm),
)
def test_unsupported_ec_operations(key, algorithm):
    """The crypto provider should raise NotImplementedError when a key doesn't support an operation or algorithm"""

    provider = get_local_cryptography_provider(key.key)
    if isinstance(algorithm, EncryptionAlgorithm):
        with pytest.raises(NotImplementedError):
            provider.encrypt(algorithm, b"...")
        with pytest.raises(NotImplementedError):
            provider.decrypt(algorithm, b"...")
    if isinstance(algorithm, KeyWrapAlgorithm):
        with pytest.raises(NotImplementedError):
            provider.wrap_key(algorithm, b"...")
        with pytest.raises(NotImplementedError):
            provider.unwrap_key(algorithm, b"...")
    elif isinstance(algorithm, SignatureAlgorithm):
        with pytest.raises(NotImplementedError):
            provider.sign(algorithm, b"...")
        with pytest.raises(NotImplementedError):
            provider.verify(algorithm, b"...", b"...")


@pytest.mark.parametrize(
    "algorithm",
    [a for a in KeyWrapAlgorithm if a.startswith("RSA")]
    + list(SignatureAlgorithm)
    + [a for a in EncryptionAlgorithm if not a.endswith("PAD")],
)
def test_unsupported_symmetric_operations(algorithm):
    """The crypto provider should raise NotImplementedError when a key doesn't support an operation or algorithm"""

    jwk = {"k": os.urandom(32), "kty": "oct", "key_ops": ("unwrapKey", "wrapKey")}
    key = KeyVaultKey(key_id="http://localhost/keys/key/version", jwk=jwk)
    provider = get_local_cryptography_provider(key.key)
    if isinstance(algorithm, EncryptionAlgorithm):
        with pytest.raises(NotImplementedError):
            provider.encrypt(algorithm, b"...")
        with pytest.raises(NotImplementedError):
            provider.decrypt(algorithm, b"...")
    if isinstance(algorithm, KeyWrapAlgorithm):
        with pytest.raises(NotImplementedError):
            provider.wrap_key(algorithm, b"...")
        with pytest.raises(NotImplementedError):
            provider.unwrap_key(algorithm, b"...")
    elif isinstance(algorithm, SignatureAlgorithm):
        with pytest.raises(NotImplementedError):
            provider.sign(algorithm, b"...")
        with pytest.raises(NotImplementedError):
            provider.verify(algorithm, b"...", b"...")



# pylint: disable=line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import asyncio
import os
from azure.identity.aio import DefaultAzureCredential
from azure.keyvault.keys.aio import KeyClient


# ----------------------------------------------------------------------------------------------------------
# Prerequisites:
# 1. An Azure Key Vault (https://learn.microsoft.com/azure/key-vault/quick-create-cli)
#
# 2. azure-keyvault-keys and azure-identity libraries (pip install these)
#
# 3. Set environment variable VAULT_URL with the URL of your key vault
#
# 4. Set up your environment to use azure-identity's DefaultAzureCredential. For more information about how to configure
#    the DefaultAzureCredential, refer to https://aka.ms/azsdk/python/identity/docs#azure.identity.DefaultAzureCredential
#
# 5. Key create, delete, recover, and purge permissions for your service principal in your vault
#
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates deleting and purging keys
#
# 1. Create a key (create_key)
#
# 2. Delete a key (delete_key)
#
# 3. Recover a deleted key (recover_deleted_key)
#
# 4. Purge a deleted key (purge_deleted_key)
# ----------------------------------------------------------------------------------------------------------
async def run_sample():
    # Instantiate a key client that will be used to call the service.
    # Here we use the DefaultAzureCredential, but any azure-identity credential can be used.
    VAULT_URL = os.environ["VAULT_URL"]
    credential = DefaultAzureCredential()
    client = KeyClient(vault_url=VAULT_URL, credential=credential)

    print("\n.. Create keys")
    rsa_key = await client.create_rsa_key("rsaKeyNameAsync")
    ec_key = await client.create_ec_key("ecKeyNameAsync")
    print(f"Created key '{rsa_key.name}' of type '{rsa_key.key_type}'.")
    print(f"Created key '{ec_key.name}' of type '{ec_key.key_type}'.")

    print("\n.. Delete the keys")
    for key_name in (ec_key.name, rsa_key.name):
        deleted_key = await client.delete_key(key_name)
        print(f"Deleted key '{deleted_key.name}'")

    # A deleted key can only be recovered if the Key Vault is soft-delete enabled.
    print("\n.. Recover a deleted key")
    recovered_key = await client.recover_deleted_key(rsa_key.name)
    print(f"Recovered key '{recovered_key.name}'")

    # To permanently delete the key, the deleted key needs to be purged.
    await client.delete_key(recovered_key.name)

    # Keys will still purge eventually on their scheduled purge date, but calling `purge_deleted_key` immediately
    # purges.
    print("\n.. Purge keys")
    for key_name in (ec_key.name, rsa_key.name):
        await client.purge_deleted_key(key_name)
        print(f"Purged '{key_name}'")

    print("\nrun_sample done")
    await credential.close()
    await client.close()


if __name__ == "__main__":
    asyncio.run(run_sample())



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
from ._client import KeyClient

__all__ = ["KeyClient"]



# pylint: disable=line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import os
from azure.identity import DefaultAzureCredential
from azure.keyvault.keys import KeyClient

# ----------------------------------------------------------------------------------------------------------
# Prerequisites:
# 1. An Azure Key Vault (https://learn.microsoft.com/azure/key-vault/quick-create-cli)
#
# 2. azure-keyvault-keys and azure-identity libraries (pip install these)
#
# 3. Set environment variable VAULT_URL with the URL of your key vault
#
# 4. Set up your environment to use azure-identity's DefaultAzureCredential. For more information about how to configure
#    the DefaultAzureCredential, refer to https://aka.ms/azsdk/python/identity/docs#azure.identity.DefaultAzureCredential
#
# 5. Key create, list, and delete permissions for your service principal in your vault
#
# ----------------------------------------------------------------------------------------------------------
# Sample - demonstrates the basic list operations for keys
#
# 1. Create a key (create_key)
#
# 2. List keys from the Key Vault (list_keys)
#
# 3. List key versions from the Key Vault (list_properties_of_key_versions)
#
# 4. Delete keys (begin_delete_key)
#
# 5. List deleted keys from the Key Vault (list_deleted_keys)
#
# ----------------------------------------------------------------------------------------------------------

# Instantiate a key client that will be used to call the service.
# Here we use the DefaultAzureCredential, but any azure-identity credential can be used.
VAULT_URL = os.environ["VAULT_URL"]
credential = DefaultAzureCredential()
client = KeyClient(vault_url=VAULT_URL, credential=credential)

# Let's create keys with RSA and EC type. If the key
# already exists in the Key Vault, then a new version of the key is created.
print("\n.. Create Key")
rsa_key = client.create_rsa_key("rsaKeyName")
ec_key = client.create_ec_key("ecKeyName")
print(f"Key with name '{rsa_key.name}' was created of type '{rsa_key.key_type}'.")
print(f"Key with name '{ec_key.name}' was created of type '{ec_key.key_type}'.")

# You need to check the type of all the keys in the vault.
# Let's list the keys and print their key types.
# List operations don 't return the keys with their type information.
# So, for each returned key we call get_key to get the key with its type information.
print("\n.. List keys from the Key Vault")
keys = client.list_properties_of_keys()
for key in keys:
    retrieved_key = client.get_key(key.name)
    print(f"Key with name '{retrieved_key.name}' with type '{retrieved_key.key_type}' was found.")

# The rsa key size now should now be 3072, default - 2048. So you want to update the key in Key Vault to ensure
# it reflects the new key size. Calling create_rsa_key on an existing key creates a new version of the key in
# the Key Vault with the new key size.
new_key = client.create_rsa_key(rsa_key.name, size=3072)
print(f"New version was created for Key with name '{new_key.name}' with the updated size.")

# You should have more than one version of the rsa key at this time. Lets print all the versions of this key.
print("\n.. List versions of a key using its name")
key_versions = client.list_properties_of_key_versions(rsa_key.name)
for key in key_versions:
    print(f"Key '{key.name}' has version: '{key.version}'")

# Both the rsa key and ec key are not needed anymore. Let's delete those keys.
print("\n.. Delete the created keys...")
for key_name in (ec_key.name, rsa_key.name):
    client.begin_delete_key(key_name).wait()

# You can list all the deleted and non-purged keys.
print("\n.. List deleted keys from the Key Vault (requires soft-delete)")
deleted_keys = client.list_deleted_keys()
for deleted_key in deleted_keys:
    print(f"Key with name '{deleted_key.name}' has recovery id '{deleted_key.recovery_id}'")



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
from azure.keyvault.keys import KeyClient
from azure.keyvault.keys.crypto import CryptographyClient

from _shared.helpers import mock


def test_key_client_close():
    transport = mock.MagicMock()
    client = KeyClient(vault_url="https://localhost", credential=object(), transport=transport)
    client.close()
    assert transport.__enter__.call_count == 0
    assert transport.__exit__.call_count == 1


def test_key_client_context_manager():
    transport = mock.MagicMock()
    client = KeyClient(vault_url="https://localhost", credential=object(), transport=transport)

    with client:
        assert transport.__enter__.call_count == 1
    assert transport.__enter__.call_count == 1
    assert transport.__exit__.call_count == 1


def test_crypto_client_close():
    transport = mock.MagicMock()
    client = CryptographyClient(key="https://localhost/a/b/c", credential=object(), transport=transport)
    client.close()
    assert transport.__enter__.call_count == 0
    assert transport.__exit__.call_count == 1


def test_crypto_client_context_manager():
    transport = mock.MagicMock()
    client = CryptographyClient(key="https://localhost/a/b/c", credential=object(), transport=transport)

    with client:
        assert transport.__enter__.call_count == 1
    assert transport.__enter__.call_count == 1
    assert transport.__exit__.call_count == 1



# pylint: disable=line-too-long,useless-suppression
# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
import asyncio
import codecs
from dateutil import parser as date_parse
import functools
import json
import logging
import os
from unittest.mock import Mock, patch

from azure.core.exceptions import HttpResponseError, ResourceExistsError, ResourceNotFoundError
from azure.core.pipeline.policies import SansIOHTTPPolicy
from azure.core.rest import HttpRequest
from azure.keyvault.keys import (
    ApiVersion,
    JsonWebKey,
    KeyProperties,
    KeyReleasePolicy,
    KeyRotationLifetimeAction,
    KeyRotationPolicy,
    KeyRotationPolicyAction,
)
from azure.keyvault.keys.aio import KeyClient
from azure.keyvault.keys._shared.client_base import DEFAULT_VERSION
import pytest

from _shared.test_case_async import KeyVaultTestCase
from _test_case import get_decorator, get_release_policy, is_public_cloud
from _async_test_case import get_attestation_token, AsyncKeysClientPreparer
from test_key_client import _assert_lifetime_actions_equal, _assert_rotation_policies_equal
from devtools_testutils.aio import recorded_by_proxy_async
from _keys_test_case import KeysTestCase


all_api_versions = get_decorator(is_async=True)
only_hsm = get_decorator(only_hsm=True, is_async=True)
only_hsm_default = get_decorator(only_hsm=True, is_async=True, api_versions=[DEFAULT_VERSION])
only_hsm_7_4_plus = get_decorator(only_hsm=True, is_async=True, api_versions=[ApiVersion.V7_4, ApiVersion.V7_5])
only_vault_7_4_plus = get_decorator(only_vault=True, is_async=True, api_versions=[ApiVersion.V7_4, ApiVersion.V7_5])
only_7_4_plus = get_decorator(is_async=True, api_versions=[ApiVersion.V7_4, ApiVersion.V7_5])
logging_enabled = get_decorator(is_async=True, logging_enable=True)
logging_disabled = get_decorator(is_async=True, logging_enable=False)

LIST_TEST_SIZE = 7


# used for logging tests
class MockHandler(logging.Handler):
    def __init__(self):
        super(MockHandler, self).__init__()
        self.messages = []

    def emit(self, record):
        self.messages.append(record)


class TestKeyVaultKey(KeyVaultTestCase, KeysTestCase):

    def _assert_jwks_equal(self, jwk1, jwk2):
        for field in JsonWebKey._FIELDS:
            if field != "key_ops":
                assert getattr(jwk1, field) == getattr(jwk2, field)

    def _assert_key_attributes_equal(self, k1: KeyProperties, k2: KeyProperties) -> None:
        assert k1.name == k2.name
        assert k1.vault_url == k2.vault_url
        assert k1.enabled == k2.enabled
        assert k1.not_before == k2.not_before
        assert k1.expires_on == k2.expires_on
        assert k1.created_on == k2.created_on
        assert k1.updated_on == k2.updated_on
        assert k1.tags == k2.tags
        assert k1.recovery_level == k2.recovery_level
        assert k1.hsm_platform == k2.hsm_platform

    async def _create_rsa_key(self, client, key_name, **kwargs):
        key_ops = ["encrypt", "decrypt", "sign", "verify", "wrapKey", "unwrapKey"]
        hsm = kwargs.get("hardware_protected") or False
        if self.is_live:
            await asyncio.sleep(2)  # to avoid throttling by the service
        created_key = await client.create_rsa_key(key_name, **kwargs)
        key_type = "RSA-HSM" if hsm else "RSA"
        self._validate_rsa_key_bundle(created_key, client.vault_url, key_name, key_type, key_ops)
        return created_key

    async def _create_ec_key(self, client, key_name, **kwargs):
        key_curve = kwargs.get("curve") or "P-256"
        hsm = kwargs.get("hardware_protected") or False
        if self.is_live:
            await asyncio.sleep(2)  # to avoid throttling by the service
        created_key = await client.create_ec_key(key_name, **kwargs)
        key_type = "EC-HSM" if hsm else "EC"
        self._validate_ec_key_bundle(key_curve, created_key, client.vault_url, key_name, key_type)
        return created_key

    def _validate_ec_key_bundle(self, key_curve, key_attributes, vault, key_name, kty):
        prefix = "/".join(s.strip("/") for s in [vault, "keys", key_name])
        key = key_attributes.key
        kid = key_attributes.id
        assert key_curve == key.crv
        assert kid.index(prefix) == 0, f"Key Id should start with '{prefix}', but value is '{kid}'"
        assert key.kty == kty, f"kty should be '{kty}', but is '{key.kty}'"
        assert (
            key_attributes.properties.created_on and key_attributes.properties.updated_on
        ), "Missing required date attributes."

    def _validate_rsa_key_bundle(self, key_attributes, vault, key_name, kty, key_ops):
        prefix = "/".join(s.strip("/") for s in [vault, "keys", key_name])
        key = key_attributes.key
        kid = key_attributes.id
        assert kid.index(prefix) == 0, f"Key Id should start with '{prefix}', but value is '{kid}'"
        assert key.kty == kty, f"kty should be '{kty}', but is '{key.kty}'"
        assert key.n and key.e, "Bad RSA public material."
        assert sorted(key_ops) == sorted(key.key_ops), f"keyOps should be '{key_ops}', but is '{key.key_ops}'"
        assert (
            key_attributes.properties.created_on and key_attributes.properties.updated_on
        ), "Missing required date attributes."

    async def _update_key_properties(self, client, key, release_policy=None):
        expires = date_parse.parse("2050-01-02T08:00:00.000Z")
        tags = {"foo": "updated tag"}
        key_ops = ["decrypt", "encrypt"]

        # wait before updating the key to make sure updated_on has a different value
        if self.is_live:
            await asyncio.sleep(2)
        key_bundle = await client.update_key_properties(
            key.name, key_operations=key_ops, expires_on=expires, tags=tags, release_policy=release_policy
        )

        assert tags == key_bundle.properties.tags
        assert key.id == key_bundle.id
        assert key.properties.updated_on != key_bundle.properties.updated_on
        assert sorted(key_ops) == sorted(key_bundle.key_operations)
        if release_policy:
            assert key.properties.release_policy.encoded_policy != key_bundle.properties.release_policy.encoded_policy
        return key_bundle

    async def _validate_key_list(self, keys, expected):
        async for key in keys:
            if key.name in expected.keys():
                self._assert_key_attributes_equal(expected[key.name].properties, key)
                del expected[key.name]
        assert len(expected) == 0

    async def _import_test_key(self, client, name, hardware_protected=False, **kwargs):
        def _to_bytes(hex):
            if len(hex) % 2:
                hex = f"0{hex}"
            return codecs.decode(hex, "hex_codec")

        key = JsonWebKey(
            kty="RSA-HSM" if hardware_protected else "RSA",
            key_ops=["encrypt", "decrypt", "sign", "verify", "wrapKey", "unwrapKey"],
            n=_to_bytes(
                "00a0914d00234ac683b21b4c15d5bed887bdc959c2e57af54ae734e8f00720d775d275e455207e3784ceeb60a50a4655dd72a7a94d271e8ee8f7959a669ca6e775bf0e23badae991b4529d978528b4bd90521d32dd2656796ba82b6bbfc7668c8f5eeb5053747fd199319d29a8440d08f4412d527ff9311eda71825920b47b1c46b11ab3e91d7316407e89c7f340f7b85a34042ce51743b27d4718403d34c7b438af6181be05e4d11eb985d38253d7fe9bf53fc2f1b002d22d2d793fa79a504b6ab42d0492804d7071d727a06cf3a8893aa542b1503f832b296371b6707d4dc6e372f8fe67d8ded1c908fde45ce03bc086a71487fa75e43aa0e0679aa0d20efe35"
            ),
            e=_to_bytes("10001"),
            d=_to_bytes(
                "627c7d24668148fe2252c7fa649ea8a5a9ed44d75c766cda42b29b660e99404f0e862d4561a6c95af6a83d213e0a2244b03cd28576473215073785fb067f015da19084ade9f475e08b040a9a2c7ba00253bb8125508c9df140b75161d266be347a5e0f6900fe1d8bbf78ccc25eeb37e0c9d188d6e1fc15169ba4fe12276193d77790d2326928bd60d0d01d6ead8d6ac4861abadceec95358fd6689c50a1671a4a936d2376440a41445501da4e74bfb98f823bd19c45b94eb01d98fc0d2f284507f018ebd929b8180dbe6381fdd434bffb7800aaabdd973d55f9eaf9bb88a6ea7b28c2a80231e72de1ad244826d665582c2362761019de2e9f10cb8bcc2625649"
            ),
            p=_to_bytes(
                "00d1deac8d68ddd2c1fd52d5999655b2cf1565260de5269e43fd2a85f39280e1708ffff0682166cb6106ee5ea5e9ffd9f98d0becc9ff2cda2febc97259215ad84b9051e563e14a051dce438bc6541a24ac4f014cf9732d36ebfc1e61a00d82cbe412090f7793cfbd4b7605be133dfc3991f7e1bed5786f337de5036fc1e2df4cf3"
            ),
            q=_to_bytes(
                "00c3dc66b641a9b73cd833bc439cd34fc6574465ab5b7e8a92d32595a224d56d911e74624225b48c15a670282a51c40d1dad4bc2e9a3c8dab0c76f10052dfb053bc6ed42c65288a8e8bace7a8881184323f94d7db17ea6dfba651218f931a93b8f738f3d8fd3f6ba218d35b96861a0f584b0ab88ddcf446b9815f4d287d83a3237"
            ),
            dp=_to_bytes(
                "00c9a159be7265cbbabc9afcc4967eb74fe58a4c4945431902d1142da599b760e03838f8cbd26b64324fea6bdc9338503f459793636e59b5361d1e6951e08ddb089e1b507be952a81fbeaf7e76890ea4f536e25505c3f648b1e88377dfc19b4c304e738dfca07211b792286a392a704d0f444c0a802539110b7f1f121c00cff0a9"
            ),
            dq=_to_bytes(
                "00a0bd4c0a3d9f64436a082374b5caf2488bac1568696153a6a5e4cd85d186db31e2f58f024c617d29f37b4e6b54c97a1e25efec59c4d1fd3061ac33509ce8cae5c11f4cd2e83f41a8264f785e78dc0996076ee23dfdfc43d67c463afaa0180c4a718357f9a6f270d542479a0f213870e661fb950abca4a14ca290570ba7983347"
            ),
            qi=_to_bytes(
                "009fe7ae42e92bc04fcd5780464bd21d0c8ac0c599f9af020fde6ab0a7e7d1d39902f5d8fb6c614184c4c1b103fb46e94cd10a6c8a40f9991a1f28269f326435b6c50276fda6493353c650a833f724d80c7d522ba16c79f0eb61f672736b68fb8be3243d10943c4ab7028d09e76cfb5892222e38bc4d35585bf35a88cd68c73b07"
            ),
        )
        imported_key = await client.import_key(name, key, **kwargs)
        self._validate_rsa_key_bundle(imported_key, client.vault_url, name, key.kty, key.key_ops)
        return imported_key

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", all_api_versions)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_key_crud_operations(self, client, is_hsm, **kwargs):
        assert client is not None

        # create ec key
        async with client:
            ec_key_name = self.get_resource_name("crud-ec-key")
            tags = {"purpose": "unit test", "test name": "CreateECKeyTest"}
            ec_key = await self._create_ec_key(
                client, enabled=True, key_name=ec_key_name, hardware_protected=is_hsm, tags=tags
            )
            assert ec_key.properties.enabled
            assert tags == ec_key.properties.tags
            # create ec with curve
            ec_key_curve_name = self.get_resource_name("crud-P-256-ec-key")
            created_ec_key_curve = await self._create_ec_key(
                client, key_name=ec_key_curve_name, curve="P-256", hardware_protected=is_hsm
            )
            assert "P-256" == created_ec_key_curve.key.crv

            # import key
            import_test_key_name = self.get_resource_name("import-test-key")
            await self._import_test_key(client, import_test_key_name, hardware_protected=is_hsm)

            # create rsa key
            rsa_key_name = self.get_resource_name("crud-rsa-key")
            tags = {"purpose": "unit test", "test name ": "CreateRSAKeyTest"}
            key_ops = ["encrypt", "decrypt", "sign", "verify", "wrapKey", "unwrapKey"]
            rsa_key = await self._create_rsa_key(
                client, key_name=rsa_key_name, key_operations=key_ops, size=2048, tags=tags, hardware_protected=is_hsm
            )
            assert tags == rsa_key.properties.tags

            # get the created key with version
            key = await client.get_key(rsa_key.name, rsa_key.properties.version)
            assert key.properties.version == rsa_key.properties.version
            self._assert_key_attributes_equal(rsa_key.properties, key.properties)

            # get key without version
            self._assert_key_attributes_equal(rsa_key.properties, (await client.get_key(rsa_key.name)).properties)

            # update key with version
            if self.is_live:
                # wait to ensure the key's update time won't equal its creation time
                await asyncio.sleep(1)

            await self._update_key_properties(client, rsa_key)

            # delete the new key
            deleted_key = await client.delete_key(rsa_key.name)
            assert deleted_key is not None

            # aside from key_ops, the original updated keys should have the same JWKs
            self._assert_jwks_equal(rsa_key.key, deleted_key.key)
            assert deleted_key.id == rsa_key.id
            assert (
                deleted_key.recovery_id and deleted_key.deleted_date and deleted_key.scheduled_purge_date
            ), "Missing required deleted key attributes."

            # get the deleted key when soft deleted enabled
            deleted_key = await client.get_deleted_key(rsa_key.name)
            assert deleted_key is not None
            assert rsa_key.id == deleted_key.id

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_hsm)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_rsa_public_exponent(self, client, **kwargs):
        """The public exponent of a Managed HSM RSA key can be specified during creation"""
        assert client is not None

        key_name = self.get_resource_name("rsa-key")
        key = await self._create_rsa_key(client, key_name, hardware_protected=True, public_exponent=17)
        public_exponent = key.key.e[0]
        assert public_exponent == 17

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", all_api_versions)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_backup_restore(self, client, is_hsm, **kwargs):
        assert client is not None

        key_name = self.get_resource_name("keybak")

        # create key
        created_bundle = await self._create_rsa_key(client, key_name, hardware_protected=is_hsm)

        # backup key
        key_backup = await client.backup_key(created_bundle.name)
        # self.assertIsNotNone(key_backup, "key_backup")
        assert key_backup is not None

        # delete key
        await client.delete_key(created_bundle.name)

        # purge key
        await client.purge_deleted_key(created_bundle.name)

        # restore key
        restore_function = functools.partial(client.restore_key_backup, key_backup)
        restored_key = await self._poll_until_no_exception(restore_function, expected_exception=ResourceExistsError)
        self._assert_key_attributes_equal(created_bundle.properties, restored_key.properties)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", all_api_versions)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_key_list(self, client, is_hsm, **kwargs):
        assert client is not None

        max_keys = LIST_TEST_SIZE
        expected = {}

        # create many keys
        for x in range(max_keys):
            key_name = self.get_resource_name(f"key{x}")
            key = await self._create_rsa_key(client, key_name, hardware_protected=is_hsm)
            expected[key.name] = key

        # list keys
        result = client.list_properties_of_keys(max_page_size=max_keys - 1)
        async for key in result:
            if key.name in expected.keys():
                self._assert_key_attributes_equal(expected[key.name].properties, key)
                del expected[key.name]
        assert len(expected) == 0

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", all_api_versions)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_list_versions(self, client, is_hsm, **kwargs):
        assert client is not None

        key_name = self.get_resource_name("testKey")

        max_keys = LIST_TEST_SIZE
        expected = {}

        # create many key versions
        for _ in range(max_keys):
            key = await self._create_rsa_key(client, key_name, hardware_protected=is_hsm)
            expected[key.id] = key

        result = client.list_properties_of_key_versions(key_name, max_page_size=max_keys - 1)

        # validate list key versions with attributes
        async for key in result:
            if key.id in expected.keys():
                expected_key = expected[key.id]
                del expected[key.id]
                self._assert_key_attributes_equal(expected_key.properties, key)
        assert 0 == len(expected)

    @pytest.mark.asyncio
    @pytest.mark.skip("Temporarily disabled due to service issue")
    @pytest.mark.parametrize("api_version,is_hsm", all_api_versions)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_list_deleted_keys(self, client, is_hsm, **kwargs):
        assert client is not None

        expected = {}

        # create keys to delete
        for i in range(LIST_TEST_SIZE):
            key_name = self.get_resource_name(f"key{i}")
            expected[key_name] = await self._create_rsa_key(client, key_name, hardware_protected=is_hsm)

        # delete all keys
        for key_name in expected.keys():
            await client.delete_key(key_name)

        # validate list deleted keys with attributes
        async for deleted_key in client.list_deleted_keys():
            assert deleted_key.deleted_date is not None
            assert deleted_key.scheduled_purge_date is not None
            assert deleted_key.recovery_id is not None

        # validate all our deleted keys are returned by list_deleted_keys
        result = client.list_deleted_keys()
        async for key in result:
            if key.name in expected.keys():
                self._assert_key_attributes_equal(expected[key.name].properties, key.properties)
                del expected[key.name]
        assert len(expected) == 0

    @pytest.mark.asyncio
    @pytest.mark.skip("Temporarily disabled due to service issue")
    @pytest.mark.parametrize("api_version,is_hsm", all_api_versions)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_recover(self, client, is_hsm, **kwargs):
        assert client is not None

        # create keys
        keys = {}
        for i in range(LIST_TEST_SIZE):
            key_name = self.get_resource_name(f"key{i}")
            keys[key_name] = await self._create_rsa_key(client, key_name, hardware_protected=is_hsm)

        # delete them
        for key_name in keys.keys():
            await client.delete_key(key_name)

        # recover them
        for key_name in keys.keys():
            recovered_key = await client.recover_deleted_key(key_name)
            expected_key = keys[key_name]
            self._assert_key_attributes_equal(expected_key.properties, recovered_key.properties)

        # validate the recovered keys
        expected = {k: v for k, v in keys.items()}
        actual = {}
        for k in expected.keys():
            actual[k] = await client.get_key(k)

        assert len(set(expected.keys()) & set(actual.keys())) == len(expected)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", all_api_versions)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_purge(self, client, is_hsm, **kwargs):
        assert client is not None

        # create keys
        key_names = [self.get_resource_name(f"key{i}") for i in range(LIST_TEST_SIZE)]
        for key_name in key_names:
            await self._create_rsa_key(client, key_name, hardware_protected=is_hsm)

        # delete them
        for key_name in key_names:
            await client.delete_key(key_name)

        # purge them
        for key_name in key_names:
            await client.purge_deleted_key(key_name)
        for key_name in key_names:
            await self._poll_until_exception(
                functools.partial(client.get_deleted_key, key_name), expected_exception=ResourceNotFoundError
            )

        # validate none are returned by list_deleted_keys
        async for deleted_key in client.list_deleted_keys():
            assert deleted_key.name not in key_names

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", logging_enabled)
    @AsyncKeysClientPreparer(logging_enable=True)
    @recorded_by_proxy_async
    async def test_logging_enabled(self, client, is_hsm, **kwargs):
        mock_handler = MockHandler()

        logger = logging.getLogger("azure")
        logger.addHandler(mock_handler)
        logger.setLevel(logging.DEBUG)

        rsa_key_name = self.get_resource_name("rsa-key-name")
        await self._create_rsa_key(client, rsa_key_name, size=2048, hardware_protected=is_hsm)

        for message in mock_handler.messages:
            if message.levelname == "DEBUG" and message.funcName == "on_request":
                # parts of the request are logged on new lines in a single message
                if "'/n" in message.message:
                    request_sections = message.message.split("/n")
                else:
                    request_sections = message.message.split("\n")
                for section in request_sections:
                    try:
                        # the body of the request should be JSON
                        body = json.loads(section)
                        expected_kty = "RSA-HSM" if is_hsm else "RSA"
                        if body["kty"] == expected_kty:
                            mock_handler.close()
                            return
                    except (ValueError, KeyError):
                        # this means the request section is not JSON or has no kty property
                        pass

        mock_handler.close()
        assert False, "Expected request body wasn't logged"

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", logging_disabled)
    @AsyncKeysClientPreparer(logging_enable=False)
    @recorded_by_proxy_async
    async def test_logging_disabled(self, client, is_hsm, **kwargs):
        mock_handler = MockHandler()

        logger = logging.getLogger("azure")
        logger.addHandler(mock_handler)
        logger.setLevel(logging.DEBUG)

        rsa_key_name = self.get_resource_name("rsa-key-name")
        await self._create_rsa_key(client, rsa_key_name, size=2048, hardware_protected=is_hsm)

        for message in mock_handler.messages:
            if message.levelname == "DEBUG" and message.funcName == "on_request":
                # parts of the request are logged on new lines in a single message
                if "'/n" in message.message:
                    request_sections = message.message.split("/n")
                else:
                    request_sections = message.message.split("\n")
                for section in request_sections:
                    try:
                        # the body of the request should be JSON
                        body = json.loads(section)
                        expected_kty = "RSA-HSM" if is_hsm else "RSA"
                        if body["kty"] == expected_kty:
                            mock_handler.close()
                            assert False, "Client request body was logged"
                    except (ValueError, KeyError):
                        # this means the request section is not JSON or has no kty property
                        pass

        mock_handler.close()

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_hsm_7_4_plus)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_get_random_bytes(self, client, **kwargs):
        assert client

        generated_random_bytes = []
        for i in range(5):
            # [START get_random_bytes]
            # get eight random bytes from a managed HSM
            random_bytes = await client.get_random_bytes(count=8)
            # [END get_random_bytes]
            assert len(random_bytes) == 8
            assert all(random_bytes != rb for rb in generated_random_bytes)
            generated_random_bytes.append(random_bytes)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_7_4_plus)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_key_release(self, client, is_hsm, **kwargs):
        if self.is_live and os.environ["KEYVAULT_SKU"] != "premium":
            pytest.skip("This test is not supported on standard SKU vaults. Follow up with service team")
        if is_hsm and client.api_version == ApiVersion.V7_5:
            pytest.skip("Currently failing on 7.5-preview.1; skipping for now")

        attestation_uri = self._get_attestation_uri()
        attestation = await get_attestation_token(attestation_uri)
        release_policy = get_release_policy(attestation_uri)

        rsa_key_name = self.get_resource_name("rsa-key-name")
        key = await self._create_rsa_key(
            client, rsa_key_name, hardware_protected=True, exportable=True, release_policy=release_policy
        )
        assert key.properties.release_policy
        assert key.properties.release_policy.encoded_policy
        assert key.properties.exportable

        try:
            release_result = await client.release_key(rsa_key_name, attestation)
            assert release_result.value
        except HttpResponseError as ex:
            # In live pipeline tests, the service can frequently throw a transient error regarding attestation
            if self.is_live and "Target environment attestation statement cannot be verified" in ex.message:
                pytest.skip("Target environment attestation statement cannot be verified. Likely transient failure.")

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_hsm_7_4_plus)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_imported_key_release(self, client, **kwargs):
        if client.api_version == ApiVersion.V7_5:
            pytest.skip("Currently failing on 7.5-preview.1; skipping for now")

        attestation_uri = self._get_attestation_uri()
        attestation = await get_attestation_token(attestation_uri)
        release_policy = get_release_policy(attestation_uri)

        imported_key_name = self.get_resource_name("imported-key-name")
        key = await self._import_test_key(
            client, imported_key_name, hardware_protected=True, exportable=True, release_policy=release_policy
        )
        assert key.properties.release_policy
        assert key.properties.release_policy.encoded_policy
        assert key.properties.exportable

        release_result = await client.release_key(imported_key_name, attestation)
        assert release_result.value

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_7_4_plus)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_update_release_policy(self, client, **kwargs):
        if self.is_live and os.environ["KEYVAULT_SKU"] != "premium":
            pytest.skip("This test is not supported on standard SKU vaults. Follow up with service team")
        if client.api_version == ApiVersion.V7_5:
            pytest.skip("Currently failing on 7.5-preview.1; skipping for now")

        attestation_uri = self._get_attestation_uri()
        release_policy = get_release_policy(attestation_uri)
        key_name = self.get_resource_name("key-name")
        key = await self._create_rsa_key(
            client, key_name, hardware_protected=True, exportable=True, release_policy=release_policy
        )

        policy = json.loads(key.properties.release_policy.encoded_policy.decode())
        claim_condition = policy["anyOf"][0]["anyOf"][0]["equals"]
        # for some reason, claim_condition may be 'true' here for KV, but should be True here for MHSM
        claim_condition = claim_condition if isinstance(claim_condition, bool) else json.loads(claim_condition)
        assert claim_condition is True

        new_release_policy_json = {
            "anyOf": [
                {"anyOf": [{"claim": "sdk-test", "equals": False}], "authority": attestation_uri.rstrip("/") + "/"}
            ],
            "version": "1.0.0",
        }
        policy_string = json.dumps(new_release_policy_json).encode()
        new_release_policy = KeyReleasePolicy(policy_string)

        updated_key = await self._update_key_properties(client, key, new_release_policy)
        updated_policy = json.loads(updated_key.properties.release_policy.encoded_policy.decode())
        claim_condition = updated_policy["anyOf"][0]["anyOf"][0]["equals"]
        claim_condition = claim_condition if isinstance(claim_condition, bool) else json.loads(claim_condition)
        assert claim_condition is False

    # Immutable policies aren't currently supported on Managed HSM
    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_vault_7_4_plus)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_immutable_release_policy(self, client, **kwargs):
        if self.is_live and os.environ["KEYVAULT_SKU"] != "premium":
            pytest.skip("This test is not supported on standard SKU vaults. Follow up with service team")

        attestation_uri = self._get_attestation_uri()
        release_policy = get_release_policy(attestation_uri, immutable=True)
        key_name = self.get_resource_name("key-name")
        key = await self._create_rsa_key(
            client, key_name, hardware_protected=True, exportable=True, release_policy=release_policy
        )
        assert key.properties.release_policy.encoded_policy
        assert key.properties.release_policy.immutable

        new_release_policy_json = {
            "anyOf": [
                {"anyOf": [{"claim": "sdk-test", "equals": False}], "authority": attestation_uri.rstrip("/") + "/"}
            ],
            "version": "1.0.0",
        }
        policy_string = json.dumps(new_release_policy_json).encode()
        new_release_policy = KeyReleasePolicy(policy_string, immutable=True)

        with pytest.raises(HttpResponseError):
            await self._update_key_properties(client, key, new_release_policy)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_7_4_plus)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_key_rotation(self, client, is_hsm, **kwargs):
        if not is_public_cloud() and self.is_live:
            pytest.skip("This test is not supported in usgov/china region. Follow up with service team.")

        key_name = self.get_resource_name("rotation-key")
        key = await self._create_rsa_key(client, key_name, hardware_protected=is_hsm)

        # MHSM doesn't automatically give keys a default rotation policy, unlike KV
        if is_hsm:
            actions = [KeyRotationLifetimeAction(KeyRotationPolicyAction.rotate, time_after_create="P6M")]
            await client.update_key_rotation_policy(
                key_name, KeyRotationPolicy(lifetime_actions=actions, expires_in="P1Y")
            )
        rotated_key = await client.rotate_key(key_name)

        # the rotated key should have a new ID, version, and key material (for RSA, n and e fields)
        assert key.id != rotated_key.id
        assert key.properties.version != rotated_key.properties.version
        assert key.key.n != rotated_key.key.n

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_7_4_plus)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_key_rotation_policy(self, client, is_hsm, **kwargs):
        if not is_public_cloud() and self.is_live:
            pytest.skip("This test is not supported in usgov/china region. Follow up with service team.")

        key_name = self.get_resource_name("rotation-key")
        await self._create_rsa_key(client, key_name, hardware_protected=is_hsm)

        # ensure passing an empty policy with no kwargs doesn't raise an error on KV (MHSM requires an expiry time)
        if not is_hsm:
            await client.update_key_rotation_policy(key_name, KeyRotationPolicy())

        # updating a rotation policy with an empty policy and override(s)
        actions = [KeyRotationLifetimeAction(KeyRotationPolicyAction.rotate, time_after_create="P2M")]
        if is_hsm:
            updated_policy = await client.update_key_rotation_policy(
                key_name, KeyRotationPolicy(), lifetime_actions=actions, expires_in="P6M"
            )
            assert updated_policy.expires_in == "P6M"
        else:  # try a policy without an expiry time (only allowed on KV)
            updated_policy = await client.update_key_rotation_policy(
                key_name, KeyRotationPolicy(), lifetime_actions=actions
            )
            assert updated_policy.expires_in is None
        fetched_policy = await client.get_key_rotation_policy(key_name)
        _assert_rotation_policies_equal(updated_policy, fetched_policy)

        updated_policy_actions = None
        for i in range(len(updated_policy.lifetime_actions)):
            if updated_policy.lifetime_actions[i].action.lower() == KeyRotationPolicyAction.rotate.lower():
                updated_policy_actions = updated_policy.lifetime_actions[i]
        assert updated_policy_actions, "Specified rotation policy action not found in updated policy"
        assert updated_policy_actions.time_after_create == "P2M"
        assert updated_policy_actions.time_before_expiry is None

        fetched_policy_actions = None
        for i in range(len(fetched_policy.lifetime_actions)):
            if fetched_policy.lifetime_actions[i].action.lower() == KeyRotationPolicyAction.rotate.lower():
                fetched_policy_actions = fetched_policy.lifetime_actions[i]
        assert fetched_policy_actions, "Specified rotation policy action not found in fetched policy"
        _assert_lifetime_actions_equal(updated_policy_actions, fetched_policy_actions)

        # updating with a round-tripped policy and overriding expires_in
        new_policy = await client.update_key_rotation_policy(key_name, policy=updated_policy, expires_in="P90D")
        assert new_policy.expires_in == "P90D"

        new_policy_actions = None
        for i in range(len(new_policy.lifetime_actions)):
            if new_policy.lifetime_actions[i].action.lower() == KeyRotationPolicyAction.rotate.lower():
                new_policy_actions = new_policy.lifetime_actions[i]
        _assert_lifetime_actions_equal(updated_policy_actions, new_policy_actions)

        # at this time, MHSM doesn't support notify actions
        if not is_hsm:
            # updating with a round-tripped policy and overriding lifetime_actions
            newest_actions = [KeyRotationLifetimeAction(KeyRotationPolicyAction.notify, time_before_expiry="P60D")]
            newest_policy = await client.update_key_rotation_policy(
                key_name, policy=new_policy, lifetime_actions=newest_actions
            )
            newest_fetched_policy = await client.get_key_rotation_policy(key_name)
            assert newest_policy.expires_in == "P90D"
            _assert_rotation_policies_equal(newest_policy, newest_fetched_policy)

            newest_policy_actions = None
            for i in range(len(newest_policy.lifetime_actions)):
                if newest_policy.lifetime_actions[i].action.lower() == KeyRotationPolicyAction.notify.lower():
                    newest_policy_actions = newest_policy.lifetime_actions[i]
            assert newest_policy_actions.time_after_create is None
            assert newest_policy_actions.time_before_expiry == "P60D"

            newest_fetched_policy_actions = None
            for i in range(len(newest_fetched_policy.lifetime_actions)):
                if newest_fetched_policy.lifetime_actions[i].action.lower() == KeyRotationPolicyAction.notify.lower():
                    newest_fetched_policy_actions = newest_fetched_policy.lifetime_actions[i]
            _assert_lifetime_actions_equal(newest_policy_actions, newest_fetched_policy_actions)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", all_api_versions)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_get_cryptography_client(self, client, is_hsm, **kwargs):
        key_name = self.get_resource_name("key-name")
        key = await self._create_rsa_key(client, key_name, hardware_protected=is_hsm)

        # try specifying the key version
        crypto_client = client.get_cryptography_client(key_name, key_version=key.properties.version)
        # both clients should use the same generated client
        assert client._client == crypto_client._client

        # the crypto client should successfully perform crypto operations
        plaintext = b"plaintext"
        result = await crypto_client.encrypt("RSA-OAEP", plaintext)
        assert result.key_id == key.id

        result = await crypto_client.decrypt(result.algorithm, result.ciphertext)
        assert result.key_id == key.id
        assert "RSA-OAEP" == result.algorithm
        assert plaintext == result.plaintext

        # try omitting the key version
        crypto_client = client.get_cryptography_client(key_name)
        # both clients should use the same generated client
        assert client._client == crypto_client._client

        # the crypto client should successfully perform crypto operations
        result = await crypto_client.encrypt("RSA-OAEP", plaintext)
        assert result.key_id == key.id

        result = await crypto_client.decrypt(result.algorithm, result.ciphertext)
        assert result.key_id == key.id
        assert "RSA-OAEP" == result.algorithm
        assert plaintext == result.plaintext

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_vault_7_4_plus)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_send_request(self, client, is_hsm, **kwargs):
        key_name = self.get_resource_name("key-name")
        key = await self._create_rsa_key(client, key_name)

        # fetch the key we just created
        request = HttpRequest(
            method="GET",
            url=f"keys/{key_name}/{key.properties.version}",
            headers={"Accept": "application/json"},
        )
        response = await client.send_request(request)
        assert response.json()["key"]["kid"] == key.id

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_hsm_default)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_get_key_attestation(self, client, **kwargs):
        # create a key
        key_name = self.get_resource_name("key-name")
        key = await self._create_rsa_key(client, key_name, hardware_protected=True)
        # attestation info shouldn't be included unless requested with get_key_attestation
        assert key.properties.attestation is None

        # fetch the key we just created
        fetched_key = await client.get_key_attestation(key_name)
        attestation = fetched_key.properties.attestation
        assert attestation is not None
        assert attestation.certificate_pem_file is not None
        assert attestation.private_key_attestation is not None
        assert attestation.public_key_attestation is not None
        assert attestation.version

        # create new key version to validate versioned fetching
        updated_key = await client.update_key_properties(key_name, tags={"tag": "value"})
        updated = await client.get_key_attestation(key_name)
        updated_attestation = updated.properties.attestation
        assert updated_attestation is not None
        assert updated_attestation.certificate_pem_file == attestation.certificate_pem_file
        assert updated_attestation.private_key_attestation == attestation.private_key_attestation
        assert updated_attestation.public_key_attestation == attestation.public_key_attestation
        assert updated_attestation.version != updated_key.properties.version

        original = await client.get_key_attestation(key_name, key.properties.version)
        original_attestation = original.properties.attestation
        assert original_attestation.version == attestation.version

    @pytest.mark.asyncio
    @pytest.mark.parametrize("api_version,is_hsm", only_vault_7_4_plus)
    @AsyncKeysClientPreparer()
    @recorded_by_proxy_async
    async def test_40x_handling(self, client, **kwargs):
        """Ensure 404 and 409 responses are raised with azure-core exceptions instead of generated KV ones"""

        # Test that 404 is raised correctly by fetching a nonexistent key
        with pytest.raises(ResourceNotFoundError):
            await client.get_key("key-that-does-not-exist")

        # Test that 409 is raised correctly (`create_key` shouldn't actually trigger this, but for raising behavior)
        async def run(*_, **__):
            return Mock(http_response=Mock(status_code=409))

        with patch.object(client._client._client._pipeline, "run", run):
            with pytest.raises(ResourceExistsError):
                await client.create_key("...", "RSA")
        await client.close()


@pytest.mark.asyncio
async def test_positive_bytes_count_required():
    client = KeyClient("...", object())
    with pytest.raises(ValueError):
        await client.get_random_bytes(count=0)
    with pytest.raises(ValueError):
        await client.get_random_bytes(count=-1)


def test_service_headers_allowed_in_logs():
    service_headers = {"x-ms-keyvault-network-info", "x-ms-keyvault-region", "x-ms-keyvault-service-version"}
    client = KeyClient("...", object())
    assert service_headers.issubset(client._client._config.http_logging_policy.allowed_header_names)


def test_custom_hook_policy():
    class CustomHookPolicy(SansIOHTTPPolicy):
        pass

    client = KeyClient("...", object(), custom_hook_policy=CustomHookPolicy())
    assert isinstance(client._client._config.custom_hook_policy, CustomHookPolicy)



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
from datetime import datetime, timezone
from typing import Optional


def raise_if_time_invalid(not_before: Optional[datetime], expires_on: Optional[datetime]) -> None:
    now = datetime.now(timezone.utc)
    if (not_before and expires_on) and not not_before <= now <= expires_on:
        raise ValueError(f"This client's key is useable only between {not_before} and {expires_on} (UTC)")
    if not_before and not_before > now:
        raise ValueError(f"This client's key is not useable until {not_before} (UTC)")
    if expires_on and expires_on <= now:
        raise ValueError(f"This client's key expired at {expires_on} (UTC)")



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding, utils

from ..algorithm import SignatureAlgorithm
from ..transform import SignatureTransform
from ..._enums import SignatureAlgorithm as KeyVaultSignatureAlgorithm


class RsaSignatureTransform(SignatureTransform):
    def __init__(self, key, padding_function, hash_algorithm):
        super(RsaSignatureTransform, self).__init__()
        self._key = key
        self._padding_function = padding_function
        self._hash_algorithm = hash_algorithm

    def sign(self, digest):
        return self._key.sign(digest, self._padding_function(digest), utils.Prehashed(self._hash_algorithm))

    def verify(self, digest, signature):
        self._key.verify(signature, digest, self._padding_function(digest), utils.Prehashed(self._hash_algorithm))


class RsaSsaPkcs1v15(SignatureAlgorithm):
    def create_signature_transform(self, key):
        return RsaSignatureTransform(key, lambda _: padding.PKCS1v15(), self._default_hash_algorithm)


class RsaSsaPss(SignatureAlgorithm):
    def create_signature_transform(self, key):
        return RsaSignatureTransform(key, self._get_padding, self._default_hash_algorithm)

    def _get_padding(self, digest):
        return padding.PSS(mgf=padding.MGF1(self._default_hash_algorithm), salt_length=len(digest))


class Ps256(RsaSsaPss):
    _name = KeyVaultSignatureAlgorithm.ps256
    _default_hash_algorithm = hashes.SHA256()


class Ps384(RsaSsaPss):
    _name = KeyVaultSignatureAlgorithm.ps384
    _default_hash_algorithm = hashes.SHA384()


class Ps512(RsaSsaPss):
    _name = KeyVaultSignatureAlgorithm.ps512
    _default_hash_algorithm = hashes.SHA512()


class Rs256(RsaSsaPkcs1v15):
    _name = KeyVaultSignatureAlgorithm.rs256
    _default_hash_algorithm = hashes.SHA256()


class Rs384(RsaSsaPkcs1v15):
    _name = KeyVaultSignatureAlgorithm.rs384
    _default_hash_algorithm = hashes.SHA384()


class Rs512(RsaSsaPkcs1v15):
    _name = KeyVaultSignatureAlgorithm.rs512
    _default_hash_algorithm = hashes.SHA512()


Ps256.register()
Ps384.register()
Ps512.register()
Rs256.register()
Rs384.register()
Rs512.register()



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------
"""Policy implementing Key Vault's challenge authentication protocol.

Normally the protocol is only used for the client's first service request, upon which:
1. The challenge authentication policy sends a copy of the request, without authorization or content.
2. Key Vault responds 401 with a header (the 'challenge') detailing how the client should authenticate such a request.
3. The policy authenticates according to the challenge and sends the original request with authorization.

The policy caches the challenge and thus knows how to authenticate future requests. However, authentication
requirements can change. For example, a vault may move to a new tenant. In such a case the policy will attempt the
protocol again.
"""

from copy import deepcopy
import time
from typing import Any, cast, Optional, Union
from urllib.parse import urlparse

from azure.core.credentials import (
    AccessToken,
    AccessTokenInfo,
    TokenCredential,
    TokenProvider,
    TokenRequestOptions,
    SupportsTokenInfo,
)
from azure.core.exceptions import ServiceRequestError
from azure.core.pipeline import PipelineRequest, PipelineResponse
from azure.core.pipeline.policies import BearerTokenCredentialPolicy
from azure.core.rest import HttpRequest, HttpResponse

from .http_challenge import HttpChallenge
from . import http_challenge_cache as ChallengeCache


def _enforce_tls(request: PipelineRequest) -> None:
    if not request.http_request.url.lower().startswith("https"):
        raise ServiceRequestError(
            "Bearer token authentication is not permitted for non-TLS protected (non-https) URLs."
        )


def _has_claims(challenge: str) -> bool:
    """Check if a challenge header contains claims.

    :param challenge: The challenge header to check.
    :type challenge: str

    :returns: True if the challenge contains claims; False otherwise.
    :rtype: bool
    """
    # Split the challenge into its scheme and parameters, then check if any parameter contains claims
    split_challenge = challenge.strip().split(" ", 1)
    return any("claims=" in item for item in split_challenge[1].split(","))


def _update_challenge(request: PipelineRequest, challenger: PipelineResponse) -> HttpChallenge:
    """Parse challenge from a challenge response, cache it, and return it.

    :param request: The pipeline request that prompted the challenge response.
    :type request: ~azure.core.pipeline.PipelineRequest
    :param challenger: The pipeline response containing the authentication challenge.
    :type challenger: ~azure.core.pipeline.PipelineResponse

    :returns: An HttpChallenge object representing the authentication challenge.
    :rtype: HttpChallenge
    """

    challenge = HttpChallenge(
        request.http_request.url,
        challenger.http_response.headers.get("WWW-Authenticate"),
        response_headers=challenger.http_response.headers,
    )
    ChallengeCache.set_challenge_for_url(request.http_request.url, challenge)
    return challenge


class ChallengeAuthPolicy(BearerTokenCredentialPolicy):
    """Policy for handling HTTP authentication challenges.

    :param credential: An object which can provide an access token for the vault, such as a credential from
        :mod:`azure.identity`
    :type credential: ~azure.core.credentials.TokenProvider
    :param str scopes: Lets you specify the type of access needed.
    """

    def __init__(self, credential: TokenProvider, *scopes: str, **kwargs: Any) -> None:
        # Pass `enable_cae` so `enable_cae=True` is always passed through self.authorize_request
        super(ChallengeAuthPolicy, self).__init__(credential, *scopes, enable_cae=True, **kwargs)
        self._credential: TokenProvider = credential
        self._token: Optional[Union["AccessToken", "AccessTokenInfo"]] = None
        self._verify_challenge_resource = kwargs.pop("verify_challenge_resource", True)
        self._request_copy: Optional[HttpRequest] = None

    def send(self, request: PipelineRequest[HttpRequest]) -> PipelineResponse[HttpRequest, HttpResponse]:
        """Authorize request with a bearer token and send it to the next policy.

        We implement this method to account for the valid scenario where a Key Vault authentication challenge is
        immediately followed by a CAE claims challenge. The base class's implementation would return the second 401 to
        the caller, but we should handle that second challenge as well (and only return any third 401 response).

        :param request: The pipeline request object
        :type request: ~azure.core.pipeline.PipelineRequest

        :return: The pipeline response object
        :rtype: ~azure.core.pipeline.PipelineResponse
        """
        self.on_request(request)
        try:
            response = self.next.send(request)
        except Exception:  # pylint:disable=broad-except
            self.on_exception(request)
            raise

        self.on_response(request, response)
        if response.http_response.status_code == 401:
            return self.handle_challenge_flow(request, response)
        return response

    def handle_challenge_flow(
        self,
        request: PipelineRequest[HttpRequest],
        response: PipelineResponse[HttpRequest, HttpResponse],
        consecutive_challenge: bool = False,
    ) -> PipelineResponse[HttpRequest, HttpResponse]:
        """Handle the challenge flow of Key Vault and CAE authentication.

        :param request: The pipeline request object
        :type request: ~azure.core.pipeline.PipelineRequest
        :param response: The pipeline response object
        :type response: ~azure.core.pipeline.PipelineResponse
        :param bool consecutive_challenge: Whether the challenge is arriving immediately after another challenge.
            Consecutive challenges can only be valid if a Key Vault challenge is followed by a CAE claims challenge.
            True if the preceding challenge was a Key Vault challenge; False otherwise.

        :return: The pipeline response object
        :rtype: ~azure.core.pipeline.PipelineResponse
        """
        self._token = None  # any cached token is invalid
        if "WWW-Authenticate" in response.http_response.headers:
            # If the previous challenge was a KV challenge and this one is too, return the 401
            claims_challenge = _has_claims(response.http_response.headers["WWW-Authenticate"])
            if consecutive_challenge and not claims_challenge:
                return response

            request_authorized = self.on_challenge(request, response)
            if request_authorized:
                # if we receive a challenge response, we retrieve a new token
                # which matches the new target. In this case, we don't want to remove
                # token from the request so clear the 'insecure_domain_change' tag
                request.context.options.pop("insecure_domain_change", False)
                try:
                    response = self.next.send(request)
                except Exception:  # pylint:disable=broad-except
                    self.on_exception(request)
                    raise

                # If consecutive_challenge == True, this could be a third consecutive 401
                if response.http_response.status_code == 401 and not consecutive_challenge:
                    # If the previous challenge wasn't from CAE, we can try this function one more time
                    if not claims_challenge:
                        return self.handle_challenge_flow(request, response, consecutive_challenge=True)
                self.on_response(request, response)
        return response

    def on_request(self, request: PipelineRequest) -> None:
        _enforce_tls(request)
        challenge = ChallengeCache.get_challenge_for_url(request.http_request.url)
        if challenge:
            # Note that if the vault has moved to a new tenant since our last request for it, this request will fail.
            if self._need_new_token:
                # azure-identity credentials require an AADv2 scope but the challenge may specify an AADv1 resource
                scope = challenge.get_scope() or challenge.get_resource() + "/.default"
                self._request_kv_token(scope, challenge)

            bearer_token = cast(Union["AccessToken", "AccessTokenInfo"], self._token).token
            request.http_request.headers["Authorization"] = f"Bearer {bearer_token}"
            return

        # else: discover authentication information by eliciting a challenge from Key Vault. Remove any request data,
        # saving it for later. Key Vault will reject the request as unauthorized and respond with a challenge.
        # on_challenge will parse that challenge, use the original request including the body, authorize the
        # request, and tell super to send it again.
        if request.http_request.content:
            self._request_copy = request.http_request
            bodiless_request = HttpRequest(
                method=request.http_request.method,
                url=request.http_request.url,
                headers=deepcopy(request.http_request.headers),
            )
            bodiless_request.headers["Content-Length"] = "0"
            request.http_request = bodiless_request

    def on_challenge(self, request: PipelineRequest, response: PipelineResponse) -> bool:
        try:
            # CAE challenges may not include a scope or tenant; cache from the previous challenge to use if necessary
            old_scope: Optional[str] = None
            old_tenant: Optional[str] = None
            cached_challenge = ChallengeCache.get_challenge_for_url(request.http_request.url)
            if cached_challenge:
                old_scope = cached_challenge.get_scope() or cached_challenge.get_resource() + "/.default"
                old_tenant = cached_challenge.tenant_id

            challenge = _update_challenge(request, response)
            # CAE challenges may not include a scope or tenant; use the previous challenge's values if necessary
            if challenge.claims and old_scope:
                challenge._parameters["scope"] = old_scope  # pylint:disable=protected-access
                challenge.tenant_id = old_tenant
            # azure-identity credentials require an AADv2 scope but the challenge may specify an AADv1 resource
            scope = challenge.get_scope() or challenge.get_resource() + "/.default"
        except ValueError:
            return False

        if self._verify_challenge_resource:
            resource_domain = urlparse(scope).netloc
            if not resource_domain:
                raise ValueError(f"The challenge contains invalid scope '{scope}'.")

            request_domain = urlparse(request.http_request.url).netloc
            if not request_domain.lower().endswith(f".{resource_domain.lower()}"):
                raise ValueError(
                    f"The challenge resource '{resource_domain}' does not match the requested domain. Pass "
                    "`verify_challenge_resource=False` to your client's constructor to disable this verification. "
                    "See https://aka.ms/azsdk/blog/vault-uri for more information."
                )

        # If we had created a request copy in on_request, use it now to send along the original body content
        if self._request_copy:
            request.http_request = self._request_copy

        # The tenant parsed from AD FS challenges is "adfs"; we don't actually need a tenant for AD FS authentication
        # For AD FS we skip cross-tenant authentication per https://github.com/Azure/azure-sdk-for-python/issues/28648
        if challenge.tenant_id and challenge.tenant_id.lower().endswith("adfs"):
            self.authorize_request(request, scope, claims=challenge.claims)
        else:
            self.authorize_request(request, scope, claims=challenge.claims, tenant_id=challenge.tenant_id)

        return True

    @property
    def _need_new_token(self) -> bool:
        now = time.time()
        refresh_on = getattr(self._token, "refresh_on", None)
        return not self._token or (refresh_on and refresh_on <= now) or self._token.expires_on - now < 300

    def _request_kv_token(self, scope: str, challenge: HttpChallenge) -> None:
        """Implementation of BearerTokenCredentialPolicy's _request_token method, but specific to Key Vault.

        :param str scope: The scope for which to request a token.
        :param challenge: The challenge for the request being made.
        :type challenge: HttpChallenge
        """
        # Exclude tenant for AD FS authentication
        exclude_tenant = challenge.tenant_id and challenge.tenant_id.lower().endswith("adfs")
        # The SupportsTokenInfo protocol needs TokenRequestOptions for token requests instead of kwargs
        if hasattr(self._credential, "get_token_info"):
            options: TokenRequestOptions = {"enable_cae": True}
            if challenge.tenant_id and not exclude_tenant:
                options["tenant_id"] = challenge.tenant_id
            self._token = cast(SupportsTokenInfo, self._credential).get_token_info(scope, options=options)
        else:
            if exclude_tenant:
                self._token = self._credential.get_token(scope, enable_cae=True)
            else:
                self._token = cast(TokenCredential, self._credential).get_token(
                    scope, tenant_id=challenge.tenant_id, enable_cae=True
                )
